#!/usr/bin/env python3
#
# TRLC - Treat Requirements Like Code
# Copyright (C) 2022 Bayerische Motoren Werke Aktiengesellschaft (BMW AG)
#
# This file is part of the TRLC Python Reference Implementation.
#
# TRLC is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TRLC is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with TRLC. If not, see <https://www.gnu.org/licenses/>.

import os
from argparse import ArgumentParser

from trlc import ast
from trlc import lint
from trlc.errors import TRLC_Error, Location, Message_Handler
from trlc.parser import Parser
from trlc.version import TRLC_VERSION


class Source_Manager:
    """Dependency and source manager for TRLC.

    This is the main entry point when using the Python API. Create an
    instance of this, register the files you want to look at, and
    finally call the process method.

    :param mh: The message handler to use
    :type mh: Message_Handler

    """
    def __init__(self, mh):
        assert isinstance(mh, Message_Handler)
        self.mh          = mh
        self.stab        = ast.Symbol_Table.create_global_table(mh)
        self.rsl_files   = {}
        self.check_files = {}
        self.trlc_files  = {}
        self.packages    = {}

    def create_parser(self, file_name):
        assert os.path.isfile(file_name)
        return Parser(mh           = self.mh,
                      stab         = self.stab,
                      file_name    = file_name)

    def register_package(self, package_name, file_name=None):
        if package_name in self.packages:
            self.packages[package_name]["file"] = file_name
        else:
            self.packages[package_name] = {"name" : package_name,
                                           "deps" : set(),
                                           "file" : file_name}

    def register_dependency(self, package_name, import_name):
        if import_name not in self.packages:
            self.register_package(import_name)
        self.packages[package_name]["deps"].add(import_name)

    def register_file(self, file_name):
        """Schedule a file for parsing.

        :param file_name: name of the file
        :type file_name: str
        :raise AssertionError: if the file does not exist
        :raise AssertionError: if the file is registed more than once
        :raise TRLC_Error: if the file is not a rsl/check/trlc file
        """
        assert os.path.isfile(file_name)

        if file_name.endswith(".rsl"):
            self.register_rsl_file(file_name)
        elif file_name.endswith(".check"):
            self.register_check_file(file_name)
        elif file_name.endswith(".trlc"):
            self.register_trlc_file(file_name)
        else:
            self.mh.error(Location(os.path.basename(file_name)),
                          "is not a rsl, check, or trlc file")

    def register_directory(self, dir_name):
        """Schedule a directory tree for parsing.

        :param dir_name: name of the directory
        :type file_name: str
        :raise AssertionError: if the directory does not exist
        :raise AssertionError: if any item in the directory is already \
        registered
        """
        assert os.path.isdir(dir_name)

        for path, _, files in os.walk(dir_name):
            for file_name in files:
                if file_name.endswith(".rsl"):
                    self.register_rsl_file(os.path.join(path, file_name))
                elif file_name.endswith(".check"):
                    self.register_check_file(os.path.join(path, file_name))
                elif file_name.endswith(".trlc"):
                    self.register_trlc_file(os.path.join(path, file_name))

    def register_rsl_file(self, file_name):
        assert os.path.isfile(file_name)
        assert file_name not in self.rsl_files

        self.rsl_files[file_name] = self.create_parser(file_name)
        self.rsl_files[file_name].parse_rsl_header()

        self.register_package(
            package_name = self.rsl_files[file_name].pkg.name,
            file_name    = file_name)
        for import_name in self.rsl_files[file_name].raw_deps:
            self.register_dependency(
                package_name = self.rsl_files[file_name].pkg.name,
                import_name  = import_name.value)

    def register_check_file(self, file_name):
        assert os.path.isfile(file_name)
        assert file_name not in self.check_files

        self.check_files[file_name] = self.create_parser(file_name)

    def register_trlc_file(self, file_name):
        assert os.path.isfile(file_name)
        assert file_name not in self.trlc_files

        self.trlc_files[file_name] = self.create_parser(file_name)

    def parse_rsl_files(self):
        # First, check that each package import is known
        for parser in self.rsl_files.values():
            for t_import in parser.raw_deps:
                parser.deps.append(parser.stab.lookup(self.mh,
                                                      t_import,
                                                      ast.Package))

        # Second, parse packages that have no unparsed
        # dependencies. Keep doing it until we parse everything or
        # until we have reached a fix point (in which case we have a
        # cycle in our dependencies).
        processed_packages = set()
        while self.packages:
            work_list = [pkg_name
                         for pkg_name, pkg in self.packages.items()
                         if pkg_name not in processed_packages and
                            pkg["deps"] <= processed_packages]

            if not work_list:
                conflicts = list(self.packages)
                for name in sorted(conflicts):
                    print("> %s depends on %s" %
                          (name,
                           ", ".join(sorted(set(self.packages[name]["deps"]) -
                                            processed_packages))))
                self.mh.error(
                    Location(list(self.packages.values())[0]["file"]),
                    "circular inheritence")

            for pkg in work_list:
                self.rsl_files[self.packages[pkg]["file"]].parse_rsl_file()
                processed_packages.add(pkg)
                del self.packages[pkg]

    def parse_check_files(self):
        ok = True
        for parser in self.check_files.values():
            try:
                parser.parse_check_file()
            except TRLC_Error:
                ok = False
        return ok

    def parse_trlc_files(self):
        ok = True
        for parser in self.trlc_files.values():
            try:
                parser.parse_trlc_file()
            except TRLC_Error:
                ok = False
        return ok

    def resolve_record_references(self):
        ok = True
        for package in self.stab.values(ast.Package):
            for obj in package.symbols.values(ast.Record_Object):
                try:
                    obj.resolve_references(self.mh)
                except TRLC_Error:
                    ok = False

        return ok

    def perform_checks(self):
        ok = True
        for package in self.stab.values(ast.Package):
            for obj in package.symbols.values(ast.Record_Object):
                try:
                    if not obj.perform_checks(self.mh):
                        ok = False
                except TRLC_Error:
                    ok = False

        return ok

    def perform_sanity_checks(self):
        ok = True
        for package in self.stab.values(ast.Package):
            for obj in package.symbols.values(ast.Record_Type):
                try:
                    lint.verify_record(self.mh, obj)
                except TRLC_Error:
                    ok = False
        return ok

    def process(self, lint_mode=False):
        """Parse all registered files.

        :param lint_mode: don't parse instances and perform static \
        analysis on the model and checks instead
        :type lint_mode: bool

        :return: a symbol table (or None if there were any errors)
        :rtype: Symbol_Table
        """
        assert isinstance(lint_mode, bool)

        # Parse RSL files (topologically sorted, in order to deal with
        # dependencies)
        try:
            self.parse_rsl_files()
        except TRLC_Error:
            return None

        # Parse check files. At this point we cannot introduce anything
        # new in terms of packages.
        ok = True
        if not self.parse_check_files():
            ok = False

        # If we run in lint mode, then we perform the checks now and then
        # stop. We do not process the TRLC files.
        if lint_mode:
            if not ok:
                return None
            if not self.perform_sanity_checks():
                return None
            print("Verified %u model(s) and check(s), and found no issues." %
                  (len(self.rsl_files) + len(self.check_files)))
            print("Note: The lint option is a placeholder and does"
                  " not do anything yet.")
            return self.stab

        # Parse TRLC files. Almost all the semantic analysis and name
        # resolution happens here, with the notable exception of resolving
        # record references (as we can have circularity here).
        if not self.parse_trlc_files():
            return None

        # Resolve record reference names and do the missing semantic
        # analysis.
        if not self.resolve_record_references():
            return None

        if not ok:
            return None

        # Finally, apply user defined checks
        if not self.perform_checks():
            return None

        return self.stab


def main():
    ap = ArgumentParser(
        description="TRLC %s (Python reference implementation)" % TRLC_VERSION
    )
    ap.add_argument("--debug-dump",
                    default=False,
                    action="store_true",
                    help="dump symbol table")
    ap.add_argument("--lint",
                    default=False,
                    action="store_true",
                    help="sanity check models and checks")
    ap.add_argument("dir_name",
                    metavar="DIR")
    options = ap.parse_args()

    mh = Message_Handler()
    sm = Source_Manager(mh)

    if not os.path.isdir(options.dir_name):
        ap.error("%s is not a directory" % options.dir_name)

    # Process input files
    ok = True
    for path, _, files in os.walk(options.dir_name):
        for file_name in files:
            if file_name.endswith(".rsl"):
                try:
                    sm.register_rsl_file(os.path.join(path, file_name))
                except TRLC_Error:
                    ok = False
            elif file_name.endswith(".check"):
                sm.register_check_file(os.path.join(path, file_name))
            elif file_name.endswith(".trlc"):
                sm.register_trlc_file(os.path.join(path, file_name))
    if not ok:
        return 1

    if sm.process(lint_mode = options.lint) is None:
        return 1
    else:
        if options.debug_dump:
            sm.stab.dump()
        return 0


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
#
# TRLC - Treat Requirements Like Code
# Copyright (C) 2022 Bayerische Motoren Werke Aktiengesellschaft (BMW AG)
#
# This file is part of the TRLC Python Reference Implementation.
#
# TRLC is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TRLC is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with TRLC. If not, see <https://www.gnu.org/licenses/>.

import os.path
import sys


class Location:
    """Reference to a source or virtual location

    Any message raised by the :class:`Message_Handler` will be
    attached to a given location. This location can be real
    (i.e. something in a file) or virtual (i.e. a builtin function).

    :attribute file_name: the name of the file or virtual location
    :type: str

    :attribute line_no: an optional line number, starting at 1
    :type: int

    :attribute col_no: an optional column number, starting at 1
    :type: int:
    """
    def __init__(self, file_name, line_no=None, col_no=None):
        assert isinstance(file_name, str)
        if line_no is not None:
            assert isinstance(line_no, int)
            assert line_no >= 1
        if col_no is not None:
            assert isinstance(col_no, int)
            assert col_no >= 1
        self.file_name   = file_name
        self.line_no     = line_no
        self.col_no      = col_no

    def to_string(self, include_column=True):
        """Return a nice string representation

        The style is the gcc-style file:line:column format. Note that
        the filename is stripped of its path in order to make the
        final message smaller.

        :param include_column: If set, include the column location (if \
        there is one)
        :type include_column: bool

        :returns: a formatted location
        :rtype: str

        """
        rv = os.path.basename(self.file_name)
        if self.line_no:
            rv += ":%u" % self.line_no
            if self.col_no and include_column:
                rv += ":%u" % self.col_no
        return rv

    def context_lines(self):
        return []


class TRLC_Error(Exception):
    """ The universal exception that TRLC raises if something goes wrong

    :attribute location: Where the issue originates from
    :type: Location

    :attribute kind: The kind of problem (e.g. lex error, error, warning, etc.)
    :type: str

    :attribute message: Description of the problem
    :type: str
    """
    def __init__(self, location, kind, message):
        assert isinstance(location, Location)
        assert isinstance(kind, str)
        assert isinstance(message, str)

        super().__init__()
        self.location = location
        self.kind     = kind
        self.message  = message


class Message_Handler:
    """Universal message handler

    All messages from TRLC are processed by this class. If you want to
    write a tool that emits additional messages then it would be a
    really good idea to also use this class. Do not use your own print
    statements.

    If the location comes from the location attribute of
    :class:`~trlc.ast.Node` then you also get context provided for
    free.

    """
    def __init__(self):
        pass

    def emit(self, location, kind, message, fatal=True):
        assert isinstance(location, Location)
        assert isinstance(kind, str)
        assert isinstance(message, str)
        assert isinstance(fatal, bool)

        context = location.context_lines()
        msg = "%s: %s: %s" % (location.to_string(len(context) == 0),
                              kind,
                              message)

        if context:
            assert len(context) == 2
            print(context[0].replace("\t", " "))
            print(context[1].replace("\t", " "), msg)
        else:
            print(msg)

        if fatal:
            raise TRLC_Error(location, kind, message)

    def lex_error(self, location, message):
        assert isinstance(location, Location)
        assert isinstance(message, str)

        self.emit(location, "lex error", message)

    def error(self, location, message, fatal=True, user=False):
        """ Create an error message

        For example::

           mh.error(my_expr.location, "potato")

        Might generate this output::

           x = 5 + 2
                 ^ foo.check:5: error: potato

        :param location: where to attach the message
        :type location: Location

        :param message: the message to print
        :type message: str

        :param fatal: should we raise an exception in addition to printing \
        the error?
        :type fatal: bool

        :param user: if set print "check error:" instead of "error:"
        :type user: bool

        :raises TRLC_Error: if fatal is true
        """

        assert isinstance(location, Location)
        assert isinstance(message, str)

        if user:
            kind = "check error"
        else:
            kind = "error"

        self.emit(location, kind, message, fatal)

    def warning(self, location, message, user=False):
        """ Create a warning message

        :param location: where to attach the message
        :type location: Location
        :param message: the message to print
        :type message: str
        :param user: if set print "check warning:" instead of "warning:"
        :type user: bool
        """
        assert isinstance(location, Location)
        assert isinstance(message, str)

        if user:
            kind = "check warning"
        else:
            kind = "warning"

        self.emit(location, kind, message, fatal=False)

    def ice_loc(self, location, message):  # pragma: no cover
        assert isinstance(location, Location)
        assert isinstance(message, str)

        self.emit(location, "ICE", message, fatal=False)
        sys.exit(1)



#!/usr/bin/env python3
#
# TRLC - Treat Requirements Like Code
# Copyright (C) 2022 Bayerische Motoren Werke Aktiengesellschaft (BMW AG)
#
# This file is part of the TRLC Python Reference Implementation.
#
# TRLC is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TRLC is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with TRLC. If not, see <https://www.gnu.org/licenses/>.

import sys

from trlc.errors import Location, Message_Handler


def triple_quoted_string_value(raw_value):
    assert isinstance(raw_value, str)
    assert raw_value.startswith("'''")
    assert raw_value.endswith("'''")

    lines = raw_value[3:-3].strip().splitlines()
    if not lines:
        return ""

    value      = lines[0]
    common_ws  = ""
    common_len = 0
    if len(lines) >= 2:
        for c in lines[1]:
            if c not in (" \t"):
                break
            common_ws  += c
            common_len += 1
    else:
        return value

    for line in lines[2:]:
        for idx in range(common_len):
            if idx < len(line) and line[idx] == common_ws[idx]:
                pass
            else:
                common_len = idx
                break

    for line in lines[1:]:
        value += "\n" + line[common_len:]

    return value


class Source_Reference(Location):
    def __init__(self, lexer, start_line, start_col, start_pos, end_pos):
        assert isinstance(lexer, Lexer)
        assert isinstance(start_line, int)
        assert isinstance(start_col, int)
        assert isinstance(start_pos, int)
        assert isinstance(end_pos, int)
        assert 0 <= start_pos <= end_pos < lexer.file_length
        super().__init__(lexer.file_name,
                         start_line,
                         start_col)
        self.lexer     = lexer
        self.start_pos = start_pos
        self.end_pos   = end_pos

    def text(self):
        return self.lexer.file_content[self.start_pos:self.end_pos + 1]

    def context_lines(self):
        line = ""
        n = self.start_pos
        while n >= 0:
            if self.lexer.file_content[n] == "\n":
                break
            line = self.lexer.file_content[n] + line
            n -= 1
        offset = self.start_pos - n - 1
        n = self.start_pos + 1
        while n < self.lexer.file_length:
            if self.lexer.file_content[n] == "\n":
                break
            line = line + self.lexer.file_content[n]
            n += 1
        maxtrail = n - self.start_pos
        tlen = self.end_pos + 1 - self.start_pos

        return [line,
                " " * offset + "^" * min(tlen, maxtrail)]


class Token:
    KIND = frozenset(["IDENTIFIER",
                      "BUILTIN",
                      "KEYWORD",
                      "BRA", "KET",
                      "S_BRA", "S_KET",
                      "C_BRA", "C_KET",
                      "COMMA",
                      "DOT",
                      "RANGE",
                      "ASSIGN",
                      "OPERATOR",
                      "ARROW",
                      "INTEGER",
                      "STRING"])

    def __init__(self, location, kind, value=None):
        assert isinstance(location, Location)
        assert kind in Token.KIND
        if kind in ("IDENTIFIER", "BUILTIN", "KEYWORD", "OPERATOR", "STRING"):
            assert isinstance(value, str)
        elif kind == "INTEGER":
            assert isinstance(value, int)
        else:
            assert value is None

        self.location = location
        self.kind     = kind
        self.value    = value

    def __repr__(self):
        if self.value is None:
            return "%s_Token" % self.kind
        else:
            return "%s_Token(%s)" % (self.kind, self.value)


class Lexer:
    KEYWORDS = frozenset(["abs",
                          "and",
                          "checks",
                          "enum",
                          "else",
                          "elsif",
                          "error",
                          "extends",
                          "false",
                          "fatal",
                          "forall",
                          "if",
                          "implies",
                          "import",
                          "in",
                          "not",
                          "null",
                          "optional",
                          "or",
                          "package",
                          "section",
                          "then",
                          "true",
                          "type",
                          "warning",
                          "xor"])

    PUNCTUATION = {
        "(" : "BRA",
        ")" : "KET",
        "{" : "C_BRA",
        "}" : "C_KET",
        "[" : "S_BRA",
        "]" : "S_KET",
        "," : "COMMA",
        "/" : "OPERATOR",
        "%" : "OPERATOR",
        "+" : "OPERATOR",
        "-" : "OPERATOR",
    }

    def __init__(self, mh, file_name):
        assert isinstance(mh, Message_Handler)
        assert isinstance(file_name, str)
        self.mh        = mh
        self.file_name = file_name

    def token(self):
        assert False


class Python_Lexer(Lexer):
    def __init__(self, mh, file_name):
        super().__init__(mh, file_name)
        with open(file_name, "r", encoding="UTF-8") as fd:
            self.file_content = fd.read()
        self.file_length  = len(self.file_content)

        self.lexpos = -2
        self.line_no = 0
        self.col_no  = 0
        self.cc = None
        self.nc = None

        self.advance()

    @staticmethod
    def is_alpha(char):
        assert isinstance(char, str) and len(char) == 1
        return ord('a') <= ord(char) <= ord('z') or \
            ord('A') <= ord(char) <= ord('Z')

    @staticmethod
    def is_numeric(char):
        assert isinstance(char, str) and len(char) == 1
        return ord('0') <= ord(char) <= ord('9')

    @staticmethod
    def is_alnum(char):
        assert isinstance(char, str) and len(char) == 1
        return ord('a') <= ord(char) <= ord('z') or \
            ord('A') <= ord(char) <= ord('Z') or \
            ord('0') <= ord(char) <= ord('9')

    def advance(self):
        self.lexpos += 1
        if self.cc == "\n" or self.lexpos == 0:
            self.line_no += 1
            self.col_no = 0
        if self.nc is not None:
            self.col_no += 1
        self.cc = self.nc
        self.nc = (self.file_content[self.lexpos + 1]
                   if self.lexpos + 1 < self.file_length
                   else None)

    def current_location(self):
        return Source_Reference(lexer      = self,
                                start_line = self.line_no,
                                start_col  = self.col_no,
                                start_pos  = self.lexpos,
                                end_pos    = self.lexpos)

    def token(self):
        # Skip whitespace and comments
        while True:
            while self.nc and self.nc.isspace():
                self.advance()
            if self.nc is None:
                return None
            self.advance()

            if self.cc == "/" and self.nc == "/":
                while self.cc and self.cc != "\n":
                    self.advance()
            elif self.cc == "/" and self.nc == "*":
                while self.nc and not (self.cc == "*" and self.nc == "/"):
                    self.advance()
                self.advance()
            else:
                break

        start_pos  = self.lexpos
        start_line = self.line_no
        start_col  = self.col_no

        if self.is_alpha(self.cc):
            kind = "IDENTIFIER"
            while self.nc and (self.is_alnum(self.nc) or
                               self.nc == "_" or
                               self.nc == ":"):
                self.advance()

        elif self.cc in Lexer.PUNCTUATION:
            kind = Lexer.PUNCTUATION[self.cc]

        elif self.cc == "=":
            if self.nc == ">":
                kind = "ARROW"
                self.advance()
            elif self.nc == "=":
                kind = "OPERATOR"
                self.advance()
            else:
                kind = "ASSIGN"

        elif self.cc == ".":
            if self.nc == ".":
                kind = "RANGE"
                self.advance()
            else:
                kind = "DOT"

        elif self.cc in ("<", ">"):
            kind = "OPERATOR"
            if self.nc == "=":
                self.advance()

        elif self.cc == "!":
            kind = "OPERATOR"
            if self.nc == "=":
                self.advance()
            else:
                self.mh.lex_error(self.current_location(),
                                  "malformed != operator")

        elif self.cc == "*":
            kind = "OPERATOR"
            if self.nc == "*":
                self.advance()

        elif self.cc == '"':
            kind = "STRING"
            while self.nc != '"':
                if self.nc is None:
                    self.mh.lex_error(
                        Source_Reference(lexer      = self,
                                         start_line = start_line,
                                         start_col  = start_col,
                                         start_pos  = start_pos,
                                         end_pos    = self.lexpos),
                        "unterminated string")
                elif self.nc == "\n":
                    self.mh.lex_error(
                        Source_Reference(lexer      = self,
                                         start_line = start_line,
                                         start_col  = start_col,
                                         start_pos  = start_pos,
                                         end_pos    = self.lexpos),
                        "double quoted strings cannot include newlines")

                self.advance()
                if self.cc == "\\" and self.nc == '"':
                    self.advance()
            self.advance()

        elif self.cc == "'":
            kind = "STRING"
            for _ in range(2):
                self.advance()
                if self.cc != "'":
                    self.mh.lex_error(
                        Source_Reference(lexer      = self,
                                         start_line = start_line,
                                         start_col  = start_col,
                                         start_pos  = start_pos,
                                         end_pos    = self.lexpos),
                        "malformed triple-quoted string")
            quotes_seen = 0
            while quotes_seen < 3:
                self.advance()
                if self.cc == "'":
                    quotes_seen += 1
                else:
                    quotes_seen = 0
                if self.nc is None:
                    self.mh.lex_error(
                        Source_Reference(lexer      = self,
                                         start_line = start_line,
                                         start_col  = start_col,
                                         start_pos  = start_pos,
                                         end_pos    = self.lexpos),
                        "unterminated triple-quoted string")

        elif self.is_numeric(self.cc):
            kind = "INTEGER"
            while self.nc and self.is_numeric(self.nc):
                self.advance()

        else:
            self.mh.lex_error(self.current_location(),
                              "unexpected character '%s'" % self.cc)

        sref = Source_Reference(lexer      = self,
                                start_line = start_line,
                                start_col  = start_col,
                                start_pos  = start_pos,
                                end_pos    = self.lexpos)

        if kind == "IDENTIFIER":
            value = sref.text()
            if value in Lexer.KEYWORDS:
                kind = "KEYWORD"
            elif ":" in value:
                kind = "BUILTIN"
                if not value.startswith("trlc:"):
                    self.mh.lex_error(sref,
                                      "builtin function name must start "
                                      "with trlc:")

        elif kind == "OPERATOR":
            value = sref.text()

        elif kind == "STRING":
            value = sref.text()
            if value.startswith('"'):
                value = value[1:-1]
                value = value.replace('\\"', '"')
            else:
                value = triple_quoted_string_value(value)

        elif kind == "INTEGER":
            value = int(sref.text())

        else:
            value = None

        return Token(sref, kind, value)


def create_lexer(mh, file_name):
    assert isinstance(mh, Message_Handler)
    assert isinstance(file_name, str)

    return Python_Lexer(mh, file_name)


def sanity_test():
    mh    = Message_Handler()
    lexer = create_lexer(mh, sys.argv[1])

    while True:
        token  = lexer.token()
        if token is None:
            break
        mh.warning(token.location,
                   str(token))


if __name__ == "__main__":
    sanity_test()



#!/usr/bin/env python3
#
# TRLC - Treat Requirements Like Code
# Copyright (C) 2022 Bayerische Motoren Werke Aktiengesellschaft (BMW AG)
#
# This file is part of the TRLC Python Reference Implementation.
#
# TRLC is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TRLC is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with TRLC. If not, see <https://www.gnu.org/licenses/>.

import re

from trlc.lexer import Token, Lexer, create_lexer
from trlc.errors import Message_Handler
from trlc import ast


class Parser:
    COMPARISON_OPERATOR = ("==", "!=", "<", "<=", ">", ">=")
    ADDING_OPERATOR = ("+", "-")
    MULTIPLYING_OPERATOR = ("*", "/", "%")

    def __init__(self, mh, stab, file_name):
        assert isinstance(mh, Message_Handler)
        assert isinstance(stab, ast.Symbol_Table)
        assert isinstance(file_name, str)
        self.mh       = mh
        self.lexer    = create_lexer(mh, file_name)
        self.stab     = stab
        self.pkg      = None
        self.raw_deps = []
        self.deps     = []
        self.imports  = set()

        self.ct = None
        self.nt = self.lexer.token()

        self.builtin_bool = stab.table["Boolean"]
        self.builtin_int  = stab.table["Integer"]
        self.builtin_str  = stab.table["String"]

        self.section = []
        self.default_scope = ast.Scope()
        self.default_scope.push(self.stab)

    def advance(self):
        self.ct = self.nt
        self.nt = self.lexer.token()

    def peek(self, kind):
        assert kind in Token.KIND
        return self.nt is not None and self.nt.kind == kind

    def peek_eof(self):
        return self.nt is None

    def peek_kw(self, value):
        assert value in Lexer.KEYWORDS
        return self.peek("KEYWORD") and self.nt.value == value

    def match(self, kind):
        assert kind in Token.KIND
        if self.nt is None:
            self.mh.error(self.ct.location,
                          "expected %s, encountered end-of-file instead" %
                          kind)
        elif self.nt.kind != kind:
            self.mh.error(self.nt.location,
                          "expected %s, encountered %s instead" %
                          (kind, self.nt.kind))
        self.advance()

    def match_eof(self):
        if self.nt is not None:
            self.mh.error(self.nt.location,
                          "expected end-of-file, encountered %s instead" %
                          self.nt.kind)

    def match_kw(self, value):
        assert value in Lexer.KEYWORDS
        if self.nt is None:
            self.mh.error(self.ct.location,
                          "expected %s, encountered end-of-file instead" %
                          value)
        elif self.nt.kind != "KEYWORD":
            self.mh.error(self.nt.location,
                          "expected %s, encountered %s instead" %
                          (value, self.nt.kind))
        elif self.nt.value != value:
            self.mh.error(self.nt.location,
                          "expected %s, encountered %s instead" %
                          (value, self.nt.value))
        self.advance()

    def parse_described_name(self):
        self.match("IDENTIFIER")
        name = self.ct

        if self.peek("STRING"):
            self.match("STRING")
            return name, self.ct.value
        else:
            return name, None

    def parse_qualified_name(self, scope, required_subclass=None):
        assert isinstance(scope, ast.Scope)
        assert isinstance(required_subclass, type)

        self.match("IDENTIFIER")
        sym = scope.lookup(self.mh, self.ct)

        if isinstance(sym, ast.Package):
            if sym != self.pkg and sym.name not in self.imports:
                self.mh.error(self.ct.location,
                              "package must be imported before use")
            self.match("DOT")
            self.match("IDENTIFIER")
            return sym.symbols.lookup(self.mh, self.ct, required_subclass)
        else:
            return scope.lookup(self.mh, self.ct, required_subclass)

    def parse_type_declaration(self):
        if self.peek_kw("enum"):
            self.parse_enum_declaration()
        else:
            self.parse_record_declaration()

    def parse_enum_declaration(self):
        self.match_kw("enum")
        name, description = self.parse_described_name()

        enum = ast.Enumeration_Type(name        = name.value,
                                    description = description,
                                    location    = name.location)
        self.pkg.symbols.register(self.mh, enum)

        self.match("C_BRA")
        while not self.peek("C_KET"):
            name, description = self.parse_described_name()
            lit = ast.Enumeration_Literal_Spec(name        = name.value,
                                               description = description,
                                               location    = name.location,
                                               enum        = enum)
            enum.literals.register(self.mh, lit)
        self.match("C_KET")

    def parse_record_declaration(self):
        self.match_kw("type")
        name, description = self.parse_described_name()

        if self.peek_kw("extends"):
            self.match_kw("extends")
            root_record = self.parse_qualified_name(self.default_scope,
                                                    ast.Record_Type)
        else:
            root_record = None

        record = ast.Record_Type(name        = name.value,
                                 description = description,
                                 location    = name.location,
                                 parent      = root_record)
        self.pkg.symbols.register(self.mh, record)

        self.match("C_BRA")
        while not self.peek("C_KET"):
            c_name, c_descr = self.parse_described_name()
            if self.peek_kw("optional"):
                self.match_kw("optional")
                c_optional = True
            else:
                c_optional = False
            c_typ = self.parse_qualified_name(self.default_scope,
                                              ast.Type)

            if self.peek("S_BRA"):
                self.match("S_BRA")
                self.match("INTEGER")
                a_lo = self.ct.value
                self.match("RANGE")
                a_loc = self.ct.location
                if self.peek("INTEGER"):
                    self.match("INTEGER")
                    a_hi = self.ct.value
                    if a_lo > a_hi:
                        self.mh.error(self.ct.location,
                                      "upper bound must be at least %u" % a_lo,
                                      fatal = False)
                    elif a_hi == 0:
                        self.mh.error(self.ct.location,
                                      "this array makes no sense",
                                      fatal = False)
                    elif a_hi == 1 and a_lo == 1:
                        self.mh.warning(a_loc,
                                        "array of fixed size 1 "
                                        "should not be an array")
                    elif a_hi == 1 and a_lo == 0:
                        self.mh.warning(a_loc,
                                        "consider making this array an"
                                        " optional %s" % c_typ.name)

                elif self.peek("OPERATOR") and self.nt.value == "*":
                    self.match("OPERATOR")
                    a_hi = None
                else:
                    self.mh.error(self.nt.location,
                                  "expected INTEGER or * for upper bound")
                self.match("S_KET")
                c_typ = ast.Array_Type(location     = a_loc,
                                       element_type = c_typ,
                                       lower_bound  = a_lo,
                                       upper_bound  = a_hi)

            comp = ast.Record_Component(name        = c_name.value,
                                        description = c_descr,
                                        location    = c_name.location,
                                        record      = record,
                                        typ         = c_typ,
                                        optional    = c_optional)
            record.components.register(self.mh, comp)

        self.match("C_KET")

    def parse_expression(self, scope):
        assert isinstance(scope, ast.Scope)

        n_lhs = self.parse_relation(scope)

        if self.peek_kw("and"):
            while self.peek_kw("and"):
                self.match_kw("and")
                t_op  = self.ct
                n_rhs = self.parse_relation(scope)
                n_lhs = ast.Binary_Expression(
                    mh       = self.mh,
                    location = t_op.location,
                    typ      = self.builtin_bool,
                    operator = ast.Binary_Operator.LOGICAL_AND,
                    n_lhs    = n_lhs,
                    n_rhs    = n_rhs)

        elif self.peek_kw("or"):
            while self.peek_kw("or"):
                self.match_kw("or")
                t_op  = self.ct
                n_rhs = self.parse_relation(scope)
                n_lhs = ast.Binary_Expression(
                    mh       = self.mh,
                    location = t_op.location,
                    typ      = self.builtin_bool,
                    operator = ast.Binary_Operator.LOGICAL_OR,
                    n_lhs    = n_lhs,
                    n_rhs    = n_rhs)

        elif self.peek_kw("xor"):
            self.match_kw("xor")
            t_op  = self.ct
            n_rhs = self.parse_relation(scope)
            n_lhs = ast.Binary_Expression(
                mh       = self.mh,
                location = t_op.location,
                typ      = self.builtin_bool,
                operator = ast.Binary_Operator.LOGICAL_XOR,
                n_lhs    = n_lhs,
                n_rhs    = n_rhs)

        elif self.peek_kw("implies"):
            self.match_kw("implies")
            t_op  = self.ct
            n_rhs = self.parse_relation(scope)
            n_lhs = ast.Binary_Expression(
                mh       = self.mh,
                location = t_op.location,
                typ      = self.builtin_bool,
                operator = ast.Binary_Operator.LOGICAL_IMPLIES,
                n_lhs    = n_lhs,
                n_rhs    = n_rhs)

        return n_lhs

    def parse_relation(self, scope):
        assert isinstance(scope, ast.Scope)
        relop_mapping = {"==" : ast.Binary_Operator.COMP_EQ,
                         "!=" : ast.Binary_Operator.COMP_NEQ,
                         "<"  : ast.Binary_Operator.COMP_LT,
                         "<=" : ast.Binary_Operator.COMP_LEQ,
                         ">"  : ast.Binary_Operator.COMP_GT,
                         ">=" : ast.Binary_Operator.COMP_GEQ}
        assert set(relop_mapping) == set(Parser.COMPARISON_OPERATOR)

        n_lhs = self.parse_simple_expression(scope)

        if self.peek("OPERATOR") and \
           self.nt.value in Parser.COMPARISON_OPERATOR:
            self.match("OPERATOR")
            t_op  = self.ct
            n_rhs = self.parse_simple_expression(scope)
            return ast.Binary_Expression(
                mh       = self.mh,
                location = t_op.location,
                typ      = self.builtin_bool,
                operator = relop_mapping[t_op.value],
                n_lhs    = n_lhs,
                n_rhs    = n_rhs)

        elif self.peek_kw("not") or self.peek_kw("in"):
            if self.peek_kw("not"):
                self.match_kw("not")
                t_not = self.ct
            else:
                t_not = None

            self.match_kw("in")
            t_in = self.ct

            n_a = self.parse_simple_expression(scope)
            if self.peek("RANGE"):
                self.match("RANGE")
                n_b = self.parse_simple_expression(scope)
                rv  = ast.Range_Test(
                    mh       = self.mh,
                    location = t_in.location,
                    typ      = self.builtin_bool,
                    n_lhs    = n_lhs,
                    n_lower  = n_a,
                    n_upper  = n_b)

            else:
                rv = ast.Binary_Expression(
                    mh       = self.mh,
                    location = t_in.location,
                    typ      = self.builtin_bool,
                    operator = ast.Binary_Operator.STRING_CONTAINS,
                    n_lhs    = n_lhs,
                    n_rhs    = n_a)

            if t_not is not None:
                rv = ast.Unary_Expression(
                    mh        = self.mh,
                    location  = t_not.location,
                    typ       = self.builtin_bool,
                    operator  = ast.Unary_Operator.LOGICAL_NOT,
                    n_operand = rv)

            return rv

        else:
            return n_lhs

    def parse_simple_expression(self, scope):
        assert isinstance(scope, ast.Scope)
        un_add_map = {"+" : ast.Unary_Operator.PLUS,
                      "-" : ast.Unary_Operator.MINUS}
        bin_add_map = {"+" : ast.Binary_Operator.PLUS,
                       "-" : ast.Binary_Operator.MINUS}
        assert set(un_add_map) == set(Parser.ADDING_OPERATOR)
        assert set(bin_add_map) == set(Parser.ADDING_OPERATOR)

        if self.peek("OPERATOR") and \
           self.nt.value in Parser.ADDING_OPERATOR:
            self.match("OPERATOR")
            t_unary = self.ct
        else:
            t_unary = None

        n_lhs = self.parse_term(scope)
        if t_unary:
            n_lhs = ast.Unary_Expression(
                mh        = self.mh,
                location  = t_unary.location,
                typ       = n_lhs.typ,
                operator  = un_add_map[t_unary.value],
                n_operand = n_lhs)

        while self.peek("OPERATOR") and \
              self.nt.value in Parser.ADDING_OPERATOR:
            self.match("OPERATOR")
            t_op  = self.ct
            n_rhs = self.parse_term(scope)
            n_lhs = ast.Binary_Expression(
                mh       = self.mh,
                location = t_op.location,
                typ      = n_lhs.typ,
                operator = bin_add_map[t_op.value],
                n_lhs    = n_lhs,
                n_rhs    = n_rhs)

        return n_lhs

    def parse_term(self, scope):
        assert isinstance(scope, ast.Scope)
        mul_map = {"*" : ast.Binary_Operator.TIMES,
                   "/" : ast.Binary_Operator.DIVIDE,
                   "%" : ast.Binary_Operator.REMAINDER}
        assert set(mul_map) == set(Parser.MULTIPLYING_OPERATOR)

        n_lhs = self.parse_factor(scope)
        while self.peek("OPERATOR") and \
              self.nt.value in Parser.MULTIPLYING_OPERATOR:
            self.match("OPERATOR")
            t_op  = self.ct
            n_rhs = self.parse_factor(scope)
            n_lhs = ast.Binary_Expression(
                mh       = self.mh,
                location = t_op.location,
                typ      = n_lhs.typ,
                operator = mul_map[t_op.value],
                n_lhs    = n_lhs,
                n_rhs    = n_rhs)

        return n_lhs

    def parse_factor(self, scope):
        assert isinstance(scope, ast.Scope)

        if self.peek_kw("not"):
            self.match_kw("not")
            t_op      = self.ct
            n_operand = self.parse_primary(scope)
            return ast.Unary_Expression(
                mh        = self.mh,
                location  = t_op.location,
                typ       = self.builtin_bool,
                operator  = ast.Unary_Operator.LOGICAL_NOT,
                n_operand = n_operand)

        elif self.peek_kw("abs"):
            self.match_kw("abs")
            t_op      = self.ct
            n_operand = self.parse_primary(scope)
            return ast.Unary_Expression(
                mh        = self.mh,
                location  = t_op.location,
                typ       = self.builtin_int,
                operator  = ast.Unary_Operator.ABSOLUTE_VALUE,
                n_operand = n_operand)

        else:
            n_lhs = self.parse_primary(scope)
            if self.peek("OPERATOR") and self.nt.value == "**":
                self.match("OPERATOR")
                t_op  = self.ct
                n_rhs = self.parse_primary(scope)
                n_rhs.evaluate(self.mh, None)
                n_lhs = ast.Binary_Expression(
                    mh       = self.mh,
                    location = t_op.location,
                    typ      = self.builtin_int,
                    operator = ast.Binary_Operator.POWER,
                    n_lhs    = n_lhs,
                    n_rhs    = n_rhs)
            return n_lhs

    def parse_primary(self, scope):
        assert isinstance(scope, ast.Scope)

        if self.peek("INTEGER"):
            self.match("INTEGER")
            return ast.Integer_Literal(self.ct, self.builtin_int)

        elif self.peek("STRING"):
            self.match("STRING")
            return ast.String_Literal(self.ct, self.builtin_str)

        elif self.peek_kw("null"):
            self.match_kw("null")
            return ast.Null_Literal(self.ct)

        elif self.peek("BRA"):
            self.match("BRA")
            if self.peek_kw("forall"):
                rv = self.parse_quantified_expression(scope)
            elif self.peek_kw("if"):
                rv = self.parse_conditional_expression(scope)
            else:
                rv = self.parse_expression(scope)
            self.match("KET")
            return rv

        else:
            return self.parse_name(scope)

    def parse_quantified_expression(self, scope):
        assert isinstance(scope, ast.Scope)

        self.match_kw("forall")
        loc = self.ct.location
        self.match("IDENTIFIER")
        t_qv = self.ct
        if scope.contains(t_qv.value):
            pdef = scope.lookup(self.mh, t_qv)
            self.mh.error(t_qv.location,
                          "shadows %s %s from %s" %
                          (pdef.__class__.__name__,
                           pdef.name,
                           pdef.location.to_string(False)))
        self.match_kw("in")
        self.match("IDENTIFIER")
        field = scope.lookup(self.mh, self.ct, ast.Record_Component)
        n_source = ast.Name_Reference(self.ct.location,
                                      field)
        if not isinstance(field.typ, ast.Array_Type):
            self.mh.error(self.ct.location,
                          "you can only quantify over arrays")
        n_var = ast.Quantified_Variable(t_qv.value,
                                        t_qv.location,
                                        field.typ.element_type)
        self.match("ARROW")

        new_table = ast.Symbol_Table()
        new_table.register(self.mh, n_var)
        scope.push(new_table)
        n_expr = self.parse_expression(scope)
        scope.pop()

        return ast.Quantified_Expression(mh         = self.mh,
                                         location   = loc,
                                         typ        = self.builtin_bool,
                                         n_variable = n_var,
                                         n_source   = n_source,
                                         n_expr     = n_expr)

    def parse_conditional_expression(self, scope):
        assert isinstance(scope, ast.Scope)

        self.match_kw("if")
        t_if = self.ct
        if_cond = self.parse_expression(scope)
        self.match_kw("then")
        if_expr = self.parse_expression(scope)
        if_action = ast.Action(self.mh, t_if, if_cond, if_expr)

        rv = ast.Conditional_Expression(location  = t_if.location,
                                        if_action = if_action)

        while self.peek_kw("elsif"):
            self.match_kw("elsif")
            t_elsif = self.ct
            elsif_cond = self.parse_expression(scope)
            self.match_kw("then")
            elsif_expr = self.parse_expression(scope)
            elsif_action = ast.Action(self.mh, t_elsif, elsif_cond, elsif_expr)
            rv.add_elsif(self.mh, elsif_action)

        self.match_kw("else")
        else_expr = self.parse_expression(scope)
        rv.set_else_part(self.mh, else_expr)

        return rv

    def parse_name(self, scope):
        # This is a bit more complex. The grammar is:
        #
        # name ::= qualified_name
        #        | qualified_name ['.' IDENTIFIER_literal]
        #        | name '[' expression ']'
        #        | BUILTIN '(' parameter_list ')'
        #
        # parameter_list ::= expression { ',' expression }

        assert isinstance(scope, ast.Scope)

        if self.peek("BUILTIN"):
            self.match("BUILTIN")
            n_name = scope.lookup(self.mh, self.ct, ast.Builtin_Function)
            t_name = self.ct

            # If the name refers to a builtin, then we have to parse
            # the arguments.
            parameters = []
            self.match("BRA")
            while not self.peek("KET"):
                parameters.append(self.parse_expression(scope))
                if self.peek("COMMA"):
                    self.match("COMMA")
                else:
                    break
            self.match("KET")
            if n_name.arity != len(parameters):
                self.mh.error(t_name.location,
                              "function %requires %u parameters" %
                              n_name.arity)

            if n_name.name == "trlc:len":
                if parameters[0].typ == self.builtin_str:
                    return ast.Unary_Expression(
                        mh        = self.mh,
                        location  = t_name.location,
                        typ       = self.builtin_int,
                        operator  = ast.Unary_Operator.STRING_LENGTH,
                        n_operand = parameters[0])
                else:
                    return ast.Unary_Expression(
                        mh        = self.mh,
                        location  = t_name.location,
                        typ       = self.builtin_int,
                        operator  = ast.Unary_Operator.ARRAY_LENGTH,
                        n_operand = parameters[0])

            elif n_name.name in ("trlc:startswith",
                                 "trlc:endswith"):
                return ast.Binary_Expression(
                    mh       = self.mh,
                    location = t_name.location,
                    typ      = self.builtin_bool,
                    operator = (ast.Binary_Operator.STRING_STARTSWITH
                                if n_name.name == "trlc:startswith"
                                else ast.Binary_Operator.STRING_ENDSWITH),
                    n_lhs    = parameters[0],
                    n_rhs    = parameters[1])

            elif n_name.name == "trlc:matches":
                parameters[1].ensure_type(self.mh, ast.Builtin_String)
                try:
                    # TODO: Fix scope for evaluate()
                    value = parameters[1].evaluate(self.mh, None)
                    assert isinstance(value.typ, ast.Builtin_String)
                    re.compile(value.value)
                except re.error as err:
                    self.mh.error(value.location,
                                  str(err))
                return ast.Binary_Expression(
                    mh       = self.mh,
                    location = t_name.location,
                    typ      = self.builtin_bool,
                    operator = ast.Binary_Operator.STRING_REGEX,
                    n_lhs    = parameters[0],
                    n_rhs    = parameters[1])

            else:
                self.mh.ice_loc(t_name.location,
                                "unexpected builtin")

        n_name = self.parse_qualified_name(scope, ast.Entity)
        t_name = self.ct

        if isinstance(n_name, ast.Enumeration_Type):
            # If the (qualified) name refers to a enum, then we have
            # to narrow it down to a literal.
            self.match("DOT")
            self.match("IDENTIFIER")
            lit = n_name.literals.lookup(self.mh,
                                         self.ct,
                                         ast.Enumeration_Literal_Spec)
            return ast.Enumeration_Literal(location = self.ct.location,
                                           literal  = lit)

        # Otherwise we have just a boring name. We do need to
        # ensure it's referring to something sane though.
        if not isinstance(n_name, (ast.Record_Component,
                                   ast.Quantified_Variable)):
            self.mh.error(t_name.location,
                          "is a %s, expected record component or "
                          "quantified variable" %
                          n_name.__class__.__name__)
        n_lhs = ast.Name_Reference(location = t_name.location,
                                   entity   = n_name)

        # Finally, if there are [] brackets, then we generate an array
        # index.
        if self.peek("S_BRA"):
            self.match("S_BRA")
            t_bracket = self.ct
            n_index = self.parse_expression(scope)
            self.match("S_KET")
            if not isinstance(n_name.typ, ast.Array_Type):
                self.mh.error(n_name.location,
                              "is not of an array type")
            n_lhs = ast.Binary_Expression(
                mh       = self.mh,
                location = t_bracket.location,
                typ      = n_name.typ.element_type,
                operator = ast.Binary_Operator.INDEX,
                n_lhs    = n_lhs,
                n_rhs    = n_index)

        return n_lhs

    def parse_check_block(self):
        self.match_kw("checks")
        self.match("IDENTIFIER")
        record = self.pkg.symbols.lookup(self.mh, self.ct, ast.Record_Type)
        scope = ast.Scope()
        scope.push(self.stab)
        scope.push(self.pkg.symbols)
        scope.push(record.components)
        self.match("C_BRA")
        while not self.peek("C_KET"):
            c_expr = self.parse_expression(scope)

            self.match("COMMA")
            if self.peek("KEYWORD"):
                self.match("KEYWORD")
                if self.ct.value not in ("warning", "error", "fatal"):
                    self.mh.error(self.ct.location,
                                  "expected warning|error|fatal")
                c_sev = self.ct.value
            else:
                c_sev = "error"

            self.match("STRING")
            t_msg = self.ct

            if self.peek("COMMA"):
                self.match("COMMA")
                self.match("IDENTIFIER")
                c_anchor = record.components.lookup(self.mh,
                                                    self.ct,
                                                    ast.Record_Component)
            else:
                c_anchor = None

            n_check = ast.Check(n_record  = record,
                                n_expr    = c_expr,
                                n_anchor  = c_anchor,
                                severity  = c_sev,
                                t_message = t_msg)
            record.add_check(n_check)

            assert scope.size() == 3

        self.match("C_KET")

    def parse_section_declaration(self):
        self.match_kw("section")
        self.match("STRING")
        if self.section:
            sec = ast.Section(name     = self.ct.value,
                              location = self.ct.location,
                              parent   = self.section[-1])
        else:
            sec = ast.Section(name     = self.ct.value,
                              location = self.ct.location,
                              parent   = None)
        self.section.append(sec)
        self.match("C_BRA")
        while not self.peek("C_KET"):
            self.parse_trlc_entry()
        self.match("C_KET")
        self.section.pop()

    def parse_boolean(self):
        self.match("KEYWORD")
        if self.ct.value in ("true", "false"):
            return ast.Boolean_Literal(self.ct, self.builtin_bool)
        else:
            self.mh.error(self.ct.location,
                          "expected boolean literal (true or false)")

    def parse_value(self, typ):
        assert isinstance(typ, ast.Type)

        if isinstance(typ, ast.Builtin_Integer):
            self.match("INTEGER")
            return ast.Integer_Literal(self.ct, self.builtin_int)

        elif isinstance(typ, ast.Builtin_String):
            self.match("STRING")
            return ast.String_Literal(self.ct, self.builtin_str)

        elif isinstance(typ, ast.Builtin_Boolean):
            return self.parse_boolean()

        elif isinstance(typ, ast.Array_Type):
            self.match("S_BRA")
            rv = ast.Array_Aggregate(self.ct.location,
                                     typ)
            while not self.peek("S_KET"):
                rv.append(self.parse_value(typ.element_type))
                if self.peek("COMMA"):
                    self.match("COMMA")
            self.match("S_KET")
            # TODO: Check length

            return rv

        elif isinstance(typ, ast.Enumeration_Type):
            enum = self.parse_qualified_name(self.default_scope,
                                             ast.Enumeration_Type)
            if enum != typ:
                self.mh.error(self.ct.location,
                              "expected %s" % typ.name)
            self.match("DOT")
            self.match("IDENTIFIER")
            lit = enum.literals.lookup(self.mh,
                                       self.ct,
                                       ast.Enumeration_Literal_Spec)
            return ast.Enumeration_Literal(self.ct.location,
                                           lit)

        elif isinstance(typ, ast.Record_Type):
            # TODO: Defer name resolution?
            # TODO: Type check
            self.match("IDENTIFIER")
            t_name = self.ct
            if self.peek("DOT"):
                self.match("DOT")
                self.match("IDENTIFIER")
                the_pkg = self.stab.lookup(self.mh, t_name, ast.Package)
                if the_pkg != self.pkg and the_pkg.name not in self.imports:
                    self.mh.error(self.ct.location,
                                  "package must be imported before use")
                t_name = self.ct
            else:
                the_pkg = self.pkg

            rv = ast.Record_Reference(t_name, typ, the_pkg)
            if the_pkg.symbols.contains(t_name.value):
                rv.target = the_pkg.symbols.lookup(self.mh,
                                                   t_name,
                                                   ast.Record_Object)
                if rv.target.e_typ != typ:
                    self.mh.error(t_name.location,
                                  "incorrect type, expected %s but %s is %s" %
                                  (typ.name,
                                   rv.name,
                                   rv.target.e_typ.name))

            return rv

        else:
            self.mh.ice_loc(self.ct.location,
                            "logic error: unexpected type %s" %
                            typ.__class__.__name__)

    def parse_record_object_declaration(self):
        r_typ = self.parse_qualified_name(self.default_scope,
                                          ast.Record_Type)
        self.match("IDENTIFIER")
        obj = ast.Record_Object(
            name     = self.ct.value,
            location = self.ct.location,
            e_typ    = r_typ,
            section  = self.section[-1] if self.section else None)
        self.pkg.symbols.register(self.mh, obj)
        self.match("C_BRA")
        while not self.peek("C_KET"):
            self.match("IDENTIFIER")
            comp = r_typ.components.lookup(self.mh,
                                           self.ct,
                                           ast.Record_Component)
            self.match("ASSIGN")
            value = self.parse_value(comp.typ)
            obj.assign(comp, value)

        # Check that each non-optional component has been specified
        for comp in r_typ.all_components():
            if obj.field[comp.name] is None:
                if not comp.optional:
                    self.mh.error(
                        obj.location,
                        "required component %s (see %s) is not defined" %
                        (comp.name,
                         comp.location.to_string(False)))
                else:
                    obj.assign(comp, ast.Implicit_Null(obj, comp))

        self.match("C_KET")

    def parse_trlc_entry(self):
        if self.peek_kw("section"):
            self.parse_section_declaration()
        else:
            self.parse_record_object_declaration()

    def parse_rsl_header(self):
        self.match_kw("package")
        self.match("IDENTIFIER")

        self.pkg = ast.Package(self.ct.value,
                               self.ct.location,
                               self.stab)
        self.stab.register(self.mh, self.pkg)
        self.default_scope.push(self.pkg.symbols)

        while self.peek_kw("import"):
            self.match_kw("import")
            self.match("IDENTIFIER")
            self.raw_deps.append(self.ct)
            self.imports.add(self.ct.value)

    def parse_rsl_file(self):
        assert self.pkg is not None

        while self.peek_kw("enum") or self.peek_kw("type"):
            self.parse_type_declaration()

        self.match_eof()

    def parse_check_file(self):
        self.match_kw("package")
        self.match("IDENTIFIER")

        self.pkg = self.stab.lookup(self.mh, self.ct, ast.Package)
        self.default_scope.push(self.pkg.symbols)

        while not self.peek_eof():
            self.parse_check_block()

        self.match_eof()

    def parse_trlc_file(self):
        self.match_kw("package")
        self.match("IDENTIFIER")

        if self.stab.contains(self.ct.value):
            self.pkg = self.stab.lookup(self.mh, self.ct, ast.Package)
        else:
            self.pkg = ast.Package(self.ct.value,
                                   self.ct.location,
                                   self.stab)
            self.pkg.declared_late = True
            self.stab.register(self.mh, self.pkg)
        self.default_scope.push(self.pkg.symbols)

        while self.peek_kw("import"):
            self.match_kw("import")
            self.match("IDENTIFIER")
            pkg = self.stab.lookup(self.mh, self.ct, ast.Package)
            if pkg.declared_late:
                self.mh.error(self.ct.location,
                              "package must be declared in rsl file to be"
                              " importable")
            self.imports.add(pkg.name)

        while self.peek_kw("section") or self.peek("IDENTIFIER"):
            self.parse_trlc_entry()

        self.match_eof()



#!/usr/bin/env python3

import sys
import setuptools

from trlc import version

with open("README.md", "r") as fd:
    long_description = fd.read()

gh_root = "https://github.com"
gh_project = "bmw-software-engineering/trlc"

project_urls = {
    "Bug Tracker"   : "%s/%s/issues" % (gh_root, gh_project),
    "Documentation" : "%s/pages/%s/" % (gh_root, gh_project),
    "Source Code"   : "%s/%s"        % (gh_root, gh_project),
}

setuptools.setup(
    name="trlc",
    version=version.TRLC_VERSION,
    author="Bayerische Motoren Werke Aktiengesellschaft (BMW AG)",
    author_email="florian.schanda@bmw.de",
    description="Treat Requirements Like Code",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=project_urls["Source Code"],
    project_urls=project_urls,
    license="GNU General Public License v3",
    packages=setuptools.find_packages(),
    # install_requires=["cvc5>=1.0.1"],
    python_requires=">=3.7, <4",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Topic :: Documentation",
        "Topic :: Software Development",
    ],
    entry_points={
        "console_scripts": [
            "trlc = trlc.trlc:main",
        ],
    },
)



#!/usr/bin/env python3
#
# TRLC - Treat Requirements Like Code
# Copyright (C) 2022 Bayerische Motoren Werke Aktiengesellschaft (BMW AG)
#
# This file is part of the TRLC Python Reference Implementation.
#
# TRLC is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TRLC is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with TRLC. If not, see <https://www.gnu.org/licenses/>.

VERSION_TUPLE = (1, 0, 4)
VERSION_SUFFIX = ""

TRLC_VERSION = ("%u.%u.%u" % VERSION_TUPLE) + \
    ("-%s" % VERSION_SUFFIX if VERSION_SUFFIX else "")

FULL_NAME = "TRLC %s" % TRLC_VERSION



#!/usr/bin/env python3
#
# TRLC - Treat Requirements Like Code
# Copyright (C) 2022 Bayerische Motoren Werke Aktiengesellschaft (BMW AG)
#
# This file is part of the TRLC Python Reference Implementation.
#
# TRLC is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TRLC is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with TRLC. If not, see <https://www.gnu.org/licenses/>.

import re

from copy import copy
from difflib import get_close_matches
from enum import Enum, auto

from trlc.errors import Location, Message_Handler
from trlc.lexer import Token
from trlc import math

#
# This module defines the AST and related object for the TRLC
# reference implementation. There are four sections:
#
# - Valuations deal with concrete values for record objects
# - AST expressions deal with the syntax tree
# - AST entities deal with concrete objects that have been declared
# - Symbol_Table and scope deals with name resolution
#


##############################################################################
# Valuations
##############################################################################

class Value:
    """Polymorphic value for evaluating expressions.

    Any record references will be fully resolved.

    :attribute location: source location this value comes from
    :type: Location

    :attribute value: the value or None (for null values)
    :type: str, int, bool, list[Value], Record_Reference, \
    Enumeration_Literal_Spec

    :attribute typ: type of the value (or None for null values)
    :type: Type
    """
    def __init__(self, location, value, typ):
        assert isinstance(location, Location)
        assert value is None or \
            isinstance(value, (str,
                               int,
                               bool,
                               list,
                               Record_Reference,
                               Enumeration_Literal_Spec))
        assert typ is None or isinstance(typ, Type)
        assert (typ is None) == (value is None)

        self.location = location
        self.value    = value
        self.typ      = typ

    def __repr__(self):  # pragma: no cover
        return "Value(%s)" % self.value

    def resolve_references(self, mh):
        assert isinstance(mh, Message_Handler)

        if isinstance(self.value, Record_Reference):
            self.value.resolve(mh)


##############################################################################
# AST Nodes
##############################################################################

class Node:
    """Base class for all AST items.

    :attribute location: source location
    :type: Location
    """
    def __init__(self, location):
        assert isinstance(location, Location)
        self.location = location

    def write_indent(self, indent, message):  # pragma: no cover
        assert isinstance(indent, int)
        assert indent >= 0
        assert isinstance(message, str)
        print(" " * (3 * indent) + message)

    def dump(self, indent=0):  # pragma: no cover
        """Visualise the parse tree.

        This can be called for any :class:`Node` or
        :class:`Symbol_Table`, and can be very helpful for debugging
        or understanding the parse tree. The dump methods will produce
        output like this::

            Symbol_Table
               Builtin_Boolean
               Builtin_Integer
               Builtin_String
               Package bar
                  Symbol_Table
                     Record_Type MyType
                        Record_Component name
                           Optional: False
                           Type: String
                        Checks
                           Error 'description is too short'
                              Anchor: description
                              Binary Binary_Operator.COMP_GT Expression
                                 Type: Boolean
                                 Unary Unary_Operator.STRING_LENGTH Expression
                                    Type: Integer
                                    Name Reference to description
                                 Integer Literal 10
               Package instances
                  Symbol_Table
                     Record_Object SomeThing
                        Type: MyType
                        Field description: "Potato"
               Builtin_Function trlc:endswith
               Builtin_Function trlc:len
               Builtin_Function trlc:matches
               Builtin_Function trlc:startswith

        """
        assert isinstance(indent, int) and indent >= 0
        assert False, "dump not implemented for %s" % self.__class__.__name__


class Check(Node):
    """User defined check

    This represent a single user-defined check inside a check block::

      checks T {
          a /= null implies a > 5, warning "potato", a
          ^^^^^^^^^^^^^^^^^^^^^^^1 ^2      ^3        ^4

    :attribute n_record: The record type this check applies to
    :type: Record_Type

    :attribute n_expr: The boolean expression for the check (see 1)
    :type: Expression

    :attribute n_anchor: The (optional) record component where the message \
    should be issued (or None) (see 4)
    :type: Record_Component

    :attribute severity: warning, error, or fatal (see 2; also if this is \
    not specified the default is 'error')
    :type: str

    :attribute message: the user-supplied message (see 3)
    :type: str
    """
    def __init__(self, n_record, n_expr, n_anchor, severity, t_message):
        assert isinstance(n_record, Record_Type)
        assert isinstance(n_expr, Expression)
        assert isinstance(n_anchor, Record_Component) or n_anchor is None
        assert severity in ("warning", "error", "fatal")
        assert isinstance(t_message, Token)
        assert t_message.kind == "STRING"
        super().__init__(t_message.location)

        self.n_record = n_record
        self.n_expr   = n_expr
        self.n_anchor = n_anchor
        self.severity = severity
        self.message  = t_message.value

    def dump(self, indent=0):  # pragma: no cover
        if self.severity == "warning":
            self.write_indent(indent, "Warning '%s'" % self.message)
        elif self.severity == "error":
            self.write_indent(indent, "Error '%s'" % self.message)
        else:
            self.write_indent(indent, "Fatal error '%s'" % self.message)
        if self.n_anchor:
            self.write_indent(indent + 1, "Anchor: %s" % self.n_anchor.name)
        self.n_expr.dump(indent + 1)

    def get_real_location(self, record_object):
        assert isinstance(record_object, Record_Object)
        if self.n_anchor is None:
            return record_object.location
        elif record_object.field[self.n_anchor.name] is not None:
            return record_object.field[self.n_anchor.name].location
        else:
            return record_object.location

    def perform(self, mh, record_object):
        assert isinstance(mh, Message_Handler)
        assert isinstance(record_object, Record_Object)

        result = self.n_expr.evaluate(mh, copy(record_object.field))
        assert isinstance(result.value, bool)

        if not result.value:
            loc = self.get_real_location(record_object)
            if self.severity == "warning":
                mh.warning(loc,
                           self.message,
                           user = True)
            else:
                mh.error(loc,
                         self.message,
                         fatal = self.severity == "fatal",
                         user  = True)
                return False

        return True

##############################################################################
# AST Nodes (Expressions)
##############################################################################


class Unary_Operator(Enum):
    MINUS          = auto()
    PLUS           = auto()
    LOGICAL_NOT    = auto()
    ABSOLUTE_VALUE = auto()

    STRING_LENGTH  = auto()
    ARRAY_LENGTH   = auto()


class Binary_Operator(Enum):
    LOGICAL_AND     = auto()    # Short-circuit
    LOGICAL_OR      = auto()    # Short-circuit
    LOGICAL_XOR     = auto()
    LOGICAL_IMPLIES = auto()    # Short-circuit

    COMP_EQ  = auto()
    COMP_NEQ = auto()
    COMP_LT  = auto()
    COMP_LEQ = auto()
    COMP_GT  = auto()
    COMP_GEQ = auto()

    STRING_CONTAINS   = auto()
    STRING_STARTSWITH = auto()
    STRING_ENDSWITH   = auto()
    STRING_REGEX      = auto()

    PLUS      = auto()
    MINUS     = auto()
    TIMES     = auto()
    DIVIDE    = auto()
    REMAINDER = auto()

    POWER = auto()

    INDEX = auto()


class Expression(Node):
    """Abstract base class for all expressions.

    :attribute typ: The type of this expression (or None for null values)
    :type: Type
    """
    def __init__(self, location, typ):
        super().__init__(location)
        assert typ is None or isinstance(typ, Type)
        self.typ = typ

    def evaluate(self, mh, context):  # pragma: no cover
        """Evaluate the expression in the given context

        The context can be None, in which case the expression is
        evaluated in a static context. Otherwise it must be a
        dictionary that maps names (such as record fields or
        quantified variables) to expressions.

        :param mh: the message handler to use
        :type mh: Message_Handler
        :param context: name mapping or None (for a static context)
        :type context: dict[str] Expression
        :raise TRLC_Error: if the expression cannot be evaluated
        :return: result of the evaluation
        :rtype: Value
        """
        assert isinstance(mh, Message_Handler)
        assert context is None or isinstance(context, dict)
        assert False, "evaluate not implemented for %s" % \
            self.__class__.__name__

    def to_string(self):  # pragma: no cover
        assert False, "to_string not implemented for %s" % \
            self.__class__.__name__

    def ensure_type(self, mh, typ):
        assert isinstance(typ, type)
        if not isinstance(self.typ, typ):
            mh.error(self.location,
                     "expected expression of type %s, got %s instead" %
                     (typ.__name__,
                      self.typ.__class__.__name__))

    def resolve_references(self, mh):
        assert isinstance(mh, Message_Handler)


class Implicit_Null(Expression):
    """Synthesised null values

    When a record object is declared and an optional component is not
    specified, we synthesise an implicit null expression for this.

    For example given this TRLC type::

      type T {
         x optional Integer
      }

    And this declaration::

      T Potato {}

    Then the field mapping for Potato will be::

      {x: Implicit_Null}

    Each field will get its own implicit null. Further note that this
    implicit null is distinct from the explicit :class:`Null_Literal`
    that can appear in check expressions.

    """
    def __init__(self, record_object, record_component):
        assert isinstance(record_object, Record_Object)
        assert isinstance(record_component, Record_Component)
        super().__init__(record_object.location, None)

    def to_string(self):
        return "null"

    def evaluate(self, mh, context):
        assert isinstance(mh, Message_Handler)
        assert context is None or isinstance(context, dict)
        return Value(self.location, None, None)

    def to_python_object(self):
        return None


class Literal(Expression):
    """Abstract base for all Literals

    Does not offer any additional features, but it's a nice way to
    group together all literal types. This is useful if you want to
    check if you are dealing with a literal::

      isinstance(my_expression, Literal)

    """
    def to_python_object(self):
        assert False


class Null_Literal(Literal):
    """The null literal

    This can appear in check expressions::

      a /= null implies a > 5
           ^^^^

    Please note that this is distinct from the :class:`Implicit_Null`
    values that appear in record objects.

    """
    def __init__(self, token):
        assert isinstance(token, Token)
        assert token.kind == "KEYWORD"
        assert token.value == "null"
        super().__init__(token.location, None)

    def dump(self, indent=0):  # pragma: no cover
        self.write_indent(indent, "Null Literal")

    def to_string(self):
        return "null"

    def evaluate(self, mh, context):
        assert isinstance(mh, Message_Handler)
        assert context is None or isinstance(context, dict)
        return Value(self.location, None, None)

    def to_python_object(self):
        return None


class Integer_Literal(Literal):
    """Integer literals

    Note that these are always positive. A negative integer is
    actually a unary negation expression, operating on a positive
    integer literal::

      x == -5

    This would create the following tree::

       OP_EQUALITY
          NAME_REFERENCE x
          UNARY_EXPRESSION -
             INTEGER_LITERAL 5

    :attribute value: the non-negative integer value
    :type: int
    """
    def __init__(self, token, typ):
        assert isinstance(token, Token)
        assert token.kind == "INTEGER"
        assert isinstance(typ, Builtin_Integer)
        super().__init__(token.location, typ)

        self.value = token.value

    def dump(self, indent=0):  # pragma: no cover
        self.write_indent(indent, "Integer Literal %u" % self.value)

    def to_string(self):
        return str(self.value)

    def evaluate(self, mh, context):
        assert isinstance(mh, Message_Handler)
        assert context is None or isinstance(context, dict)
        return Value(self.location, self.value, self.typ)

    def to_python_object(self):
        return self.value


class String_Literal(Literal):
    """String literals

    Note the value of the string does not include the quotation marks,
    and any escape sequences are fully resolved. For example::

       "foo\\"bar"

    Will have a value of ``foo"bar``.

    :attribute value: string content
    :type: str

    """
    def __init__(self, token, typ):
        assert isinstance(token, Token)
        assert token.kind == "STRING"
        assert isinstance(typ, Builtin_String)
        super().__init__(token.location, typ)

        self.value = token.value

    def dump(self, indent=0):  # pragma: no cover
        self.write_indent(indent, "String Literal %s" % self.value)

    def to_string(self):
        return self.value

    def evaluate(self, mh, context):
        assert isinstance(mh, Message_Handler)
        assert context is None or isinstance(context, dict)
        return Value(self.location, self.value, self.typ)

    def to_python_object(self):
        return self.value


class Boolean_Literal(Literal):
    """Boolean values

    :attribute value: the boolean value
    :type: bool
    """
    def __init__(self, token, typ):
        assert isinstance(token, Token)
        assert token.kind == "KEYWORD"
        assert token.value in ("false", "true")
        assert isinstance(typ, Builtin_Boolean)
        super().__init__(token.location, typ)

        self.value = token.value == "true"

    def dump(self, indent=0):  # pragma: no cover
        self.write_indent(indent, "Boolean Literal %s" % self.value)

    def to_string(self):
        return str(self.value)

    def evaluate(self, mh, context):
        assert isinstance(mh, Message_Handler)
        assert context is None or isinstance(context, dict)
        return Value(self.location, self.value, self.typ)

    def to_python_object(self):
        return self.value


class Enumeration_Literal(Literal):
    """Enumeration values

    Note that this is distinct from
    :class:`Enumeration_Literal_Spec`. An enumeration literal is a
    specific mention of an enumeration member in an expression::

       foo != my_enum.POTATO
              ^^^^^^^^^^^^^^

    To get to the string value of the enumeration literal
    (i.e. ``POTATO`` here) you can get the name of the literal spec
    itself: ``enum_lit.value.name``; and to get the name of the
    enumeration (i.e. ``my_enum`` here) you can use
    ``enum_lit.value.enum.name``.

    :attribute value: enumeration value
    :type: Enumeration_Literal_Spec

    """
    def __init__(self, location, literal):
        assert isinstance(literal, Enumeration_Literal_Spec)
        super().__init__(location, literal.enum)

        self.value = literal

    def dump(self, indent=0):  # pragma: no cover
        self.write_indent(indent,
                          "Enumeration Literal %s.%s" % (self.typ.name,
                                                         self.value.name))

    def to_string(self):
        return self.typ.name + "." + self.value.name

    def evaluate(self, mh, context):
        assert isinstance(mh, Message_Handler)
        assert context is None or isinstance(context, dict)
        return Value(self.location, self.value, self.typ)

    def to_python_object(self):
        return self.value.name


class Array_Aggregate(Expression):
    """Instances of array types

    This is created when assigning to array components::

       potatoes = ["picasso", "yukon gold", "sweet"]
                  ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    The type of expression that can be found in an array is somewhat
    limited:

    * :class:`Literal`
    * :class:`Array_Aggregate`
    * :class:`Record_Reference`

    :attribute value: contents of the array
    :type: list[Expression]

    """
    def __init__(self, location, typ):
        super().__init__(location, typ)
        self.value = []

    def append(self, value):
        assert isinstance(value, (Literal,
                                  Array_Aggregate,
                                  Record_Reference))
        self.value.append(value)

    def to_string(self):
        return "[" + ", ".join(x.to_string() for x in self.value) + "]"

    def evaluate(self, mh, context):
        assert isinstance(mh, Message_Handler)
        assert context is None or isinstance(context, dict)
        return Value(self.location,
                     list(element.evaluate(mh, context)
                          for element in self.value),
                     self.typ)

    def resolve_references(self, mh):
        assert isinstance(mh, Message_Handler)

        for val in self.value:
            val.resolve_references(mh)

    def to_python_object(self):
        return [x.to_python_object() for x in self.value]


class Record_Reference(Expression):
    """Reference to another record object

    This can appear in record object declarations::

      Requirement Kitten {
          depends_on = Other_Package.Cat
                       ^1            ^2
      }

    Note that this is distinct from :class:`Record_Object`. It is just
    the name; to get to the object referred to by this you can consult
    the target attribute.

    The reason we have this indirection is that not all names can be
    immediately resolved on parsing in the TRLC language.

    Note that while the containing package (see 1) is optional in the
    source language, the containing package will always be filled in
    in this AST node.

    :attribute name: The name of the record (see 2)
    :type: str

    :attribute target: The concrete record object referred to by (2)
    :type: Record_Object

    :attribute package: The package (see 1) supposed to contain (2)
    :type: Package

    """
    def __init__(self, referencing_token, typ, package):
        assert isinstance(referencing_token, Token)
        assert referencing_token.kind == "IDENTIFIER"
        assert isinstance(typ, Record_Type)
        assert isinstance(package, Package)
        super().__init__(referencing_token.location, typ)

        self.name    = referencing_token.value
        self.target  = None
        self.package = package

    def dump(self, indent=0):  # pragma: no cover
        self.write_indent(indent, "Record Reference %s" % self.name)
        self.write_indent(indent + 1, "Resolved: %s" % self.target is not None)

    def to_string(self):
        return self.name

    def evaluate(self, mh, context):
        assert isinstance(mh, Message_Handler)
        assert context is None or isinstance(context, dict)
        return Value(self.location, self, self.typ)

    def resolve_references(self, mh):
        assert isinstance(mh, Message_Handler)

        self.target = self.package.symbols.lookup_direct(
            mh                = mh,
            name              = self.name,
            error_location    = self.location,
            required_subclass = Record_Object)

    def to_python_object(self):
        return self.name


class Name_Reference(Expression):
    """Reference to a name

    Name reference to either a :class:`Record_Component` or a
    :class:`Quantified_Variable`. The actual value of course depends
    on the context. See :py:meth:`Expression.evaluate()`.

    For example::

      (forall x in potato => x > 1)
                   ^1        ^2

    Both indicated parts are a :class:`Name_Reference`, the first one
    refers to a :class:`Record_Component`, and the second refers to a
    :class:`Quantified_Variable`.

    :attribute entity: the entity named here
    :type: Record_Component, Quantified_Variable
    """
    def __init__(self, location, entity):
        assert isinstance(entity, (Record_Component,
                                   Quantified_Variable))
        super().__init__(location, entity.typ)
        self.entity = entity

    def dump(self, indent=0):  # pragma: no cover
        self.write_indent(indent, "Name Reference to %s" % self.entity.name)

    def to_string(self):
        return self.entity.name

    def evaluate(self, mh, context):
        assert isinstance(mh, Message_Handler)
        assert context is None or isinstance(context, dict)

        if context is None:
            mh.error(self.location,
                     "cannot be used in a static context")

        assert self.entity.name in context
        return context[self.entity.name].evaluate(mh, context)


class Unary_Expression(Expression):
    """Expression with only one operand

    This captures the following operations:

    * Unary_Operator.PLUS (e.g. ``+5``)
    * Unary_Operator.MINUS (e.g. ``-5``)
    * Unary_Operator.ABSOLUTE_VALUE (e.g. ``abs 42``)
    * Unary_Operator.LOGICAL_NOT (e.g. ``not True``)
    * Unary_Operator.STRING_LENGTH (e.g. ``trlc::len("foobar")``)
    * Unary_Operator.ARRAY_LENGTH (e.g. ``trlc::len(component_name)``)

    Note that several builtin functions are mapped to unary operators.

    :attribute operator: the operation
    :type: Unary_Operator

    :attribute n_operand: the expression we operate on
    :type: Expression

    """
    def __init__(self, mh, location, typ, operator, n_operand):
        super().__init__(location, typ)
        assert isinstance(mh, Message_Handler)
        assert isinstance(operator, Unary_Operator)
        assert isinstance(n_operand, Expression)
        self.operator  = operator
        self.n_operand = n_operand

        if operator in (Unary_Operator.MINUS,
                        Unary_Operator.PLUS,
                        Unary_Operator.ABSOLUTE_VALUE):
            self.n_operand.ensure_type(mh, Builtin_Integer)
        elif operator == Unary_Operator.LOGICAL_NOT:
            self.n_operand.ensure_type(mh, Builtin_Boolean)
        elif operator == Unary_Operator.STRING_LENGTH:
            self.n_operand.ensure_type(mh, Builtin_String)
        elif operator == Unary_Operator.ARRAY_LENGTH:
            self.n_operand.ensure_type(mh, Array_Type)
        else:
            mh.ice_loc(self.location,
                       "unexpected unary operation %s" % operator)

    def dump(self, indent=0):  # pragma: no cover
        self.write_indent(indent, "Unary %s Expression" % self.operator)
        self.write_indent(indent + 1, "Type: %s" % self.typ.name)
        self.n_operand.dump(indent + 1)

    def evaluate(self, mh, context):
        assert isinstance(mh, Message_Handler)
        assert context is None or isinstance(context, dict)

        v_operand = self.n_operand.evaluate(mh, context)
        if v_operand.value is None:
            mh.error(v_operand.location,
                     "input to unary expression %s (%s) must not be null" %
                     (self.to_string(),
                      self.location.to_string(False)))

        if self.operator == Unary_Operator.MINUS:
            return Value(location = self.location,
                         value    = -v_operand.value,
                         typ      = self.typ)
        elif self.operator == Unary_Operator.PLUS:
            return Value(location = self.location,
                         value    = +v_operand.value,
                         typ      = self.typ)
        elif self.operator == Unary_Operator.LOGICAL_NOT:
            return Value(location = self.location,
                         value    = not v_operand.value,
                         typ      = self.typ)
        elif self.operator == Unary_Operator.ABSOLUTE_VALUE:
            return Value(location = self.location,
                         value    = abs(v_operand.value),
                         typ      = self.typ)
        elif self.operator in (Unary_Operator.STRING_LENGTH,
                               Unary_Operator.ARRAY_LENGTH):
            return Value(location = self.location,
                         value    = len(v_operand.value),
                         typ      = self.typ)
        else:
            mh.ice_loc(self.location,
                       "unexpected unary operation %s" % self.operator)


class Binary_Expression(Expression):
    """Expression with two operands

    This captures the following operations:

    * Binary_Operator.LOGICAL_AND (e.g. ``a and b``)
    * Binary_Operator.LOGICAL_OR (e.g. ``a or b``)
    * Binary_Operator.LOGICAL_XOR (e.g. ``a xor b``)
    * Binary_Operator.LOGICAL_IMPLIES (e.g. ``a implies b``)
    * Binary_Operator.COMP_EQ (e.g. ``a == null``)
    * Binary_Operator.COMP_NEQ (e.g. ``a != null``)
    * Binary_Operator.COMP_LT (e.g. ``1 < 2``)
    * Binary_Operator.COMP_LEQ (e.g. ``1 <= 2``)
    * Binary_Operator.COMP_GT (e.g. ``a > b``)
    * Binary_Operator.COMP_GEQ (e.g. ``a >= b``)
    * Binary_Operator.STRING_CONTAINS (e.g. ``"foo" in "foobar"``)
    * Binary_Operator.STRING_STARTSWITH (e.g. ``trlc::startswith("foo", "f")``)
    * Binary_Operator.STRING_ENDSWITH (e.g. ``trlc::endswith("foo", "o")``)
    * Binary_Operator.STRING_REGEX (e.g. ``trlc::matches("foo", ".o.``)
    * Binary_Operator.PLUS (e.g. ``42 + b`` or ``"foo" + bar``)
    * Binary_Operator.MINUS (e.g. ``a - 1``)
    * Binary_Operator.TIMES (e.g. ``2 * x``)
    * Binary_Operator.DIVIDE (e.g. ``x / 2``)
    * Binary_Operator.REMAINDER (e.g. ``x % 2``)
    * Binary_Operator.POWER (e.g. ``x ** 2``)
    * Binary_Operator.INDEX (e.g. ``foo[2]``)

    Note that several builtin functions are mapped to unary operators.

    Note also that the plus operation is supported for both integers
    and strings.

    :attribute operator: the operation
    :type: Binary_Operator

    :attribute n_lhs: the first operand
    :type: Expression

    :attribute n_lhs: the second operand
    :type: Expression

    """
    def __init__(self, mh, location, typ, operator, n_lhs, n_rhs):
        super().__init__(location, typ)
        assert isinstance(mh, Message_Handler)
        assert isinstance(operator, Binary_Operator)
        assert isinstance(n_lhs, Expression)
        assert isinstance(n_rhs, Expression)
        self.operator = operator
        self.n_lhs    = n_lhs
        self.n_rhs    = n_rhs

        if operator in (Binary_Operator.LOGICAL_AND,
                        Binary_Operator.LOGICAL_OR,
                        Binary_Operator.LOGICAL_XOR,
                        Binary_Operator.LOGICAL_IMPLIES):
            self.n_lhs.ensure_type(mh, Builtin_Boolean)
            self.n_rhs.ensure_type(mh, Builtin_Boolean)

        elif operator in (Binary_Operator.COMP_EQ,
                          Binary_Operator.COMP_NEQ):
            if (self.n_lhs.typ is None) or (self.n_rhs.typ is None):
                # We can compary anything to null (including itself)
                pass
            elif self.n_lhs.typ != self.n_rhs.typ:
                # Otherwise we can compare anything, as long as the
                # types match
                mh.error(self.location,
                         "type mismatch: %s and %s do not match" %
                         (self.n_lhs.typ.name,
                          self.n_rhs.typ.name))

        elif operator in (Binary_Operator.COMP_LT,
                          Binary_Operator.COMP_LEQ,
                          Binary_Operator.COMP_GT,
                          Binary_Operator.COMP_GEQ):
            self.n_lhs.ensure_type(mh, Builtin_Integer)
            self.n_rhs.ensure_type(mh, Builtin_Integer)

        elif operator in (Binary_Operator.STRING_CONTAINS,
                          Binary_Operator.STRING_STARTSWITH,
                          Binary_Operator.STRING_ENDSWITH,
                          Binary_Operator.STRING_REGEX):
            self.n_lhs.ensure_type(mh, Builtin_String)
            self.n_rhs.ensure_type(mh, Builtin_String)

        elif operator == Binary_Operator.PLUS:
            if isinstance(self.n_lhs.typ, Builtin_Integer):
                self.n_rhs.ensure_type(mh, Builtin_Integer)
            else:
                self.n_lhs.ensure_type(mh, Builtin_String)
                self.n_rhs.ensure_type(mh, Builtin_String)

        elif operator in (Binary_Operator.MINUS,
                          Binary_Operator.TIMES,
                          Binary_Operator.DIVIDE,
                          Binary_Operator.REMAINDER,
                          Binary_Operator.POWER):
            self.n_lhs.ensure_type(mh, Builtin_Integer)
            self.n_rhs.ensure_type(mh, Builtin_Integer)

        elif operator == Binary_Operator.INDEX:
            self.n_lhs.ensure_type(mh, Array_Type)
            self.n_rhs.ensure_type(mh, Builtin_Integer)

        else:
            mh.ice_loc(self.location,
                       "unexpected binary operation %s" % operator)

    def dump(self, indent=0):  # pragma: no cover
        self.write_indent(indent, "Binary %s Expression" % self.operator)
        self.write_indent(indent + 1, "Type: %s" % self.typ.name)
        self.n_lhs.dump(indent + 1)
        self.n_rhs.dump(indent + 1)

    def to_string(self):
        infix_operators = {
            Binary_Operator.LOGICAL_AND     : "and",
            Binary_Operator.LOGICAL_OR      : "or",
            Binary_Operator.LOGICAL_XOR     : "xor",
            Binary_Operator.LOGICAL_IMPLIES : "implies",
            Binary_Operator.COMP_EQ         : "==",
            Binary_Operator.COMP_NEQ        : "!=",
            Binary_Operator.COMP_LT         : "<",
            Binary_Operator.COMP_LEQ        : "<=",
            Binary_Operator.COMP_GT         : ">",
            Binary_Operator.COMP_GEQ        : ">=",
            Binary_Operator.STRING_CONTAINS : "in",
            Binary_Operator.PLUS            : "+",
            Binary_Operator.MINUS           : "-",
            Binary_Operator.TIMES           : "*",
            Binary_Operator.DIVIDE          : "/",
            Binary_Operator.REMAINDER       : "%",
            Binary_Operator.POWER           : "**",
        }
        string_functions = {
            Binary_Operator.STRING_STARTSWITH : "startswith",
            Binary_Operator.STRING_ENDSWITH   : "endswith",
            Binary_Operator.STRING_REGEX      : "matches",
        }

        if self.operator in infix_operators:
            return "%s %s %s" % (self.n_lhs.to_string(),
                                 infix_operators[self.operator],
                                 self.n_rhs.to_string())

        elif self.operator in string_functions:
            return "%s(%s, %s)" % (string_functions[self.operator],
                                   self.n_lhs.to_string(),
                                   self.n_rhs.to_string())

        elif self.operator == Binary_Operator.INDEX:
            return "%s[%s]" % (self.n_lhs.to_string(),
                               self.n_rhs.to_string())

        else:
            assert False

    def evaluate(self, mh, context):
        assert isinstance(mh, Message_Handler)
        assert context is None or isinstance(context, dict)

        v_lhs = self.n_lhs.evaluate(mh, context)
        if v_lhs.value is None and \
           self.operator not in (Binary_Operator.COMP_EQ,
                                 Binary_Operator.COMP_NEQ):
            mh.error(v_lhs.location,
                     "lhs of check %s (%s) must not be null" %
                     (self.to_string(),
                      self.location.to_string(False)))

        # Check for the short-circuit operators first
        if self.operator == Binary_Operator.LOGICAL_AND:
            assert isinstance(v_lhs.value, bool)
            if v_lhs.value:
                return self.n_rhs.evaluate(mh, context)
            else:
                return v_lhs

        elif self.operator == Binary_Operator.LOGICAL_OR:
            assert isinstance(v_lhs.value, bool)
            if v_lhs.value:
                return v_lhs
            else:
                return self.n_rhs.evaluate(mh, context)

        elif self.operator == Binary_Operator.LOGICAL_IMPLIES:
            assert isinstance(v_lhs.value, bool)
            if v_lhs.value:
                return self.n_rhs.evaluate(mh, context)
            else:
                return Value(location = self.location,
                             value    = True,
                             typ      = self.typ)

        # Otherwise, evaluate RHS and do the operation
        v_rhs = self.n_rhs.evaluate(mh, context)
        if v_rhs.value is None and \
           self.operator not in (Binary_Operator.COMP_EQ,
                                 Binary_Operator.COMP_NEQ):
            mh.error(v_rhs.location,
                     "rhs of check %s (%s) must not be null" %
                     (self.to_string(),
                      self.location.to_string(False)))

        if self.operator == Binary_Operator.LOGICAL_XOR:
            assert isinstance(v_lhs.value, bool)
            assert isinstance(v_rhs.value, bool)
            return Value(location = self.location,
                         value    = v_lhs.value ^ v_rhs.value,
                         typ      = self.typ)

        elif self.operator == Binary_Operator.COMP_EQ:
            return Value(location = self.location,
                         value    = v_lhs.value == v_rhs.value,
                         typ      = self.typ)

        elif self.operator == Binary_Operator.COMP_NEQ:
            return Value(location = self.location,
                         value    = v_lhs.value != v_rhs.value,
                         typ      = self.typ)

        elif self.operator in (Binary_Operator.COMP_LT,
                               Binary_Operator.COMP_LEQ,
                               Binary_Operator.COMP_GT,
                               Binary_Operator.COMP_GEQ):
            return Value(location = self.location,
                         value    = {
                             Binary_Operator.COMP_LT  : lambda l, r: l < r,
                             Binary_Operator.COMP_LEQ : lambda l, r: l <= r,
                             Binary_Operator.COMP_GT  : lambda l, r: l > r,
                             Binary_Operator.COMP_GEQ : lambda l, r: l >= r,
                         }[self.operator](v_lhs.value, v_rhs.value),
                         typ      = self.typ)

        elif self.operator == Binary_Operator.STRING_CONTAINS:
            assert isinstance(v_lhs.value, str)
            assert isinstance(v_rhs.value, str)

            return Value(location = self.location,
                         value    = v_lhs.value in v_rhs.value,
                         typ      = self.typ)

        elif self.operator == Binary_Operator.STRING_STARTSWITH:
            assert isinstance(v_lhs.value, str)
            assert isinstance(v_rhs.value, str)
            return Value(location = self.location,
                         value    = v_lhs.value.startswith(v_rhs.value),
                         typ      = self.typ)

        elif self.operator == Binary_Operator.STRING_ENDSWITH:
            assert isinstance(v_lhs.value, str)
            assert isinstance(v_rhs.value, str)
            return Value(location = self.location,
                         value    = v_lhs.value.endswith(v_rhs.value),
                         typ      = self.typ)

        elif self.operator == Binary_Operator.STRING_REGEX:
            assert isinstance(v_lhs.value, str)
            assert isinstance(v_rhs.value, str)
            return Value(location = self.location,
                         value    = re.match(v_rhs.value,
                                             v_lhs.value) is not None,
                         typ      = self.typ)

        elif self.operator == Binary_Operator.PLUS:
            assert isinstance(v_lhs.value, (int, str))
            assert isinstance(v_rhs.value, (int, str))
            return Value(location = self.location,
                         value    = v_lhs.value + v_rhs.value,
                         typ      = self.typ)

        elif self.operator == Binary_Operator.MINUS:
            assert isinstance(v_lhs.value, int)
            assert isinstance(v_rhs.value, int)
            return Value(location = self.location,
                         value    = v_lhs.value - v_rhs.value,
                         typ      = self.typ)

        elif self.operator == Binary_Operator.TIMES:
            assert isinstance(v_lhs.value, int)
            assert isinstance(v_rhs.value, int)
            return Value(location = self.location,
                         value    = v_lhs.value * v_rhs.value,
                         typ      = self.typ)

        elif self.operator == Binary_Operator.DIVIDE:
            assert isinstance(v_lhs.value, int)
            assert isinstance(v_rhs.value, int)

            if v_rhs.value == 0:
                mh.error(v_rhs.location,
                         "division by zero in %s (%s)" %
                         (self.to_string(),
                          self.location.to_string(False)))

            return Value(location = self.location,
                         value    = v_lhs.value // v_rhs.value,
                         typ      = self.typ)

        elif self.operator == Binary_Operator.REMAINDER:
            assert isinstance(v_lhs.value, int)
            assert isinstance(v_rhs.value, int)

            if v_rhs.value == 0:
                mh.error(v_rhs.location,
                         "division by zero in %s (%s)" %
                         (self.to_string(),
                          self.location.to_string(False)))

            return Value(location = self.location,
                         value    = math.remainder(v_lhs.value, v_rhs.value),
                         typ      = self.typ)

        elif self.operator == Binary_Operator.POWER:
            assert isinstance(v_lhs.value, int)
            assert isinstance(v_rhs.value, int)
            return Value(location = self.location,
                         value    = v_lhs.value ** v_rhs.value,
                         typ      = self.typ)

        elif self.operator == Binary_Operator.INDEX:
            assert isinstance(v_lhs.value, list)
            assert isinstance(v_rhs.value, int)

            if v_rhs.value < 0:
                mh.error(v_rhs.location,
                         "index cannot be less than zero in %s (%s)" %
                         (self.to_string(),
                          self.location.to_string(False)))
            elif v_rhs.value > v_lhs.typ.upper_bound:
                mh.error(v_rhs.location,
                         "index cannot be more than %u in %s (%s)" %
                         (v_lhs.typ.upper_bound,
                          self.to_string(),
                          self.location.to_string(False)))
            elif v_rhs.value > len(v_lhs):
                mh.error(v_lhs.location,
                         "array is not big enough in %s (%s)" %
                         (self.to_string(),
                          self.location.to_string(False)))

            return Value(location = self.location,
                         value    = v_lhs.value[v_rhs.value],
                         typ      = self.typ)

        else:
            mh.ice_loc(self.location,
                       "unexpected binary operator %s" % self.operator)


class Range_Test(Expression):
    """Range membership test

    For example in::

      x in 1   ..   field+1
      ^lhs ^lower   ^^^^^^^upper

    Note that none of these are guaranteed to be literals or names;
    you can have arbitrarily complex expressions here.

    :attribute n_lhs: the expression to test
    :type: Expression

    :attribute n_lower: the lower bound
    :type: Expression

    :attribute n_lower: the upper bound
    :type: Expression

    """
    def __init__(self, mh, location, typ, n_lhs, n_lower, n_upper):
        super().__init__(location, typ)
        assert isinstance(mh, Message_Handler)
        assert isinstance(n_lhs, Expression)
        assert isinstance(n_lower, Expression)
        assert isinstance(n_upper, Expression)
        self.n_lhs   = n_lhs
        self.n_lower = n_lower
        self.n_upper = n_upper

        self.n_lhs.ensure_type(mh, Builtin_Integer)
        self.n_lower.ensure_type(mh, Builtin_Integer)
        self.n_upper.ensure_type(mh, Builtin_Integer)

    def dump(self, indent=0):  # pragma: no cover
        self.write_indent(indent, "Range Test")
        self.write_indent(indent + 1, "Type: %s" % self.typ)
        self.n_lhs.dump(indent + 1)
        self.n_lower.dump(indent + 1)
        self.n_upper.dump(indent + 1)

    def evaluate(self, mh, context):
        assert isinstance(mh, Message_Handler)
        assert context is None or isinstance(context, dict)

        v_lhs = self.n_lhs.evaluate(mh, context)
        if v_lhs.value is None:
            mh.error(v_lhs.location,
                     "lhs of range check %s (%s) see must not be null" %
                     (self.to_string(),
                      self.location.to_string(False)))

        v_lower = self.n_lower.evaluate(mh, context)
        if v_lower.value is None:
            mh.error(v_lower.location,
                     "lower bound of range check %s (%s) must not be null" %
                     (self.to_string(),
                      self.location.to_string(False)))

        v_upper = self.n_upper.evaluate(mh, context)
        if v_upper.value is None:
            mh.error(v_upper.location,
                     "upper bound of range check %s (%s) must not be null" %
                     (self.to_string(),
                      self.location.to_string(False)))

        return Value(location = self.location,
                     value    = v_lower.value <= v_lhs.value <= v_upper.value,
                     typ      = self.typ)


class Action(Node):
    """An if or elseif part inside a conditional expression

    Each :class:`Conditional_Expression` is made up of a sequence of
    Actions. For example here is a single expression with two
    Actions::

      (if x == 0 then "zero" elsif x == 1 then "one" else "lots")
       ^^^^^^^^^^^^^^^^^^^^^ ^^^^^^^^^^^^^^^^^^^^^^^

    Note that the else part is not an action, it is an attribute of
    the :class:`Conditional_Expression` itself.

    :attribute kind: Either if or elseif
    :type: str

    :attribute n_cond: The boolean condition expression
    :type: Expression

    :attribute n_expr: The value if the condition evaluates to true
    :type: Expression

    """
    def __init__(self, mh, t_kind, n_condition, n_expression):
        assert isinstance(mh, Message_Handler)
        assert isinstance(t_kind, Token)
        assert t_kind.kind == "KEYWORD"
        assert t_kind.value in ("if", "elsif")
        assert isinstance(n_condition, Expression)
        assert isinstance(n_expression, Expression)
        super().__init__(t_kind.location)
        self.kind   = t_kind.value
        self.n_cond = n_condition
        self.n_expr = n_expression

        self.n_cond.ensure_type(mh, Builtin_Boolean)

    def dump(self, indent=0):  # pragma: no cover
        self.write_indent(indent, "%s Action" % self.kind.capitalize())
        self.write_indent(indent + 1, "Condition")
        self.n_cond.dump(indent + 2)
        self.write_indent(indent + 1, "Value")
        self.n_expr.dump(indent + 2)

    def to_string(self):
        return "%s %s then %s" % (self.kind,
                                  self.n_cond.to_string(),
                                  self.n_expr.to_string())


class Conditional_Expression(Expression):
    """A conditional expression

    Each :class:`Conditional_Expression` is made up of a sequence of
    one or more :class:`Action`. For example here is a single
    expression with two Actions::

      (if x == 0 then "zero" elsif x == 1 then "one" else "lots")
       ^^^^^^^^^^^^^^^^^^^^^ ^^^^^^^^^^^^^^^^^^^^^^^      ^^^^^^

    The else expression is part of the conditional expression itself.

    A conditional expression will have at least one action (the if
    action), and all other actions will be elsif actions. The else
    expression is not optional and will always be present. The types
    of all actions and the else expression will match.

    :attribute actions: a list of Actions
    :type: list[Action]

    :attribute else_expr: the else expression
    :type: Expression

    """
    def __init__(self, location, if_action):
        assert isinstance(if_action, Action)
        assert if_action.kind == "if"
        super().__init__(location, if_action.n_expr.typ)
        self.actions   = [if_action]
        self.else_expr = None

    def add_elsif(self, mh, n_action):
        assert isinstance(mh, Message_Handler)
        assert isinstance(n_action, Action)
        assert n_action.kind == "elsif"

        if n_action.n_expr.typ != self.typ:
            mh.error(n_action.n_expr.location,
                     "expression is of type %s, but must be of type %s" %
                     (n_action.n_expr.typ.name,
                      self.typ.name))

        self.actions.append(n_action)

    def set_else_part(self, mh, n_expr):
        assert isinstance(mh, Message_Handler)
        assert isinstance(n_expr, Expression)

        if n_expr.typ != self.typ:
            mh.error(n_expr.location,
                     "expression is of type %s, but must be of type %s" %
                     (n_expr.typ.name,
                      self.typ.name))

        self.else_expr = n_expr

    def dump(self, indent=0):  # pragma: no cover
        self.write_indent(indent, "Conditional expression")
        for action in self.actions:
            action.dump(indent + 1)
        self.write_indent(indent + 1, "Else")
        self.else_expr.dump(indent + 2)

    def to_string(self):
        rv = "(" + " ".join(action.to_string()
                            for action in self.actions)
        rv += " else %s" % self.else_expr.to_string()
        rv += ")"
        return rv

    def evaluate(self, mh, context):
        assert isinstance(mh, Message_Handler)
        assert context is None or isinstance(context, dict)

        for action in self.actions:
            v_cond = action.n_cond.evaluate(mh, context)
            if v_cond.value is None:
                mh.error(v_cond.location,
                         "condition of %s (%s) must not be null" %
                         (action.to_string(),
                          action.location.to_string(False)))
            if v_cond.value:
                return action.n_expr.evaluate(mh, context)

        return self.else_expr.evaluate(mh, context)


class Quantified_Expression(Expression):
    """A quantified expression

    For example::

      (forall x in array_component => x > 0)
              ^1   ^2                 ^^^^^3

    A quantified expression introduces and binds a
    :class:`Quantified_Variable` (see 1) from a specified source (see
    2). When the body (see 3) is evaluated, the name of 1 is bound to
    each component of the source in turn.

    :attribute n_var: The quantified variable (see 1)
    :type: Quantified_Variable

    :attribute n_source: The array to iterate over (see 2)
    :type: Name_Reference

    :attribute n_expr: The body of the quantifier (see 3)
    :type: Expression

    """
    def __init__(self, mh, location, typ, n_variable, n_source, n_expr):
        super().__init__(location, typ)
        assert isinstance(n_variable, Quantified_Variable)
        assert isinstance(n_expr, Expression)
        assert isinstance(n_source, Name_Reference)
        assert isinstance(typ, Builtin_Boolean)
        self.n_var    = n_variable
        self.n_expr   = n_expr
        self.n_source = n_source
        self.n_expr.ensure_type(mh, Builtin_Boolean)

    def dump(self, indent=0):  # pragma: no cover
        self.write_indent(indent, "Quantified expression")
        self.n_var.dump(indent + 1)
        self.n_expr.dump(indent + 1)

    def to_string(self):
        return "(forall %s in %s => %s)" % (self.n_var.name,
                                            self.n_source.to_string(),
                                            self.n_expr.to_string())

    def evaluate(self, mh, context):
        assert isinstance(mh, Message_Handler)
        assert context is None or isinstance(context, dict)

        if context is None:
            new_ctx = {}
        else:
            new_ctx = copy(context)

        # This is going to be a bit tricky. We essentially eliminate
        # the quantifier and substitute; for the sake of making better
        # error messages.
        assert isinstance(self.n_source.entity, Record_Component)
        array_values = context[self.n_source.entity.name]
        if isinstance(array_values, Implicit_Null):
            mh.error(array_values.location,
                     "%s in quantified expression %s (%s) "
                     "must not be null" %
                     (self.n_source.to_string(),
                      self.to_string(),
                      self.location.to_string(False)))
        else:
            assert isinstance(array_values, Array_Aggregate)

        ok = True
        loc = self.location
        for binding in array_values.value:
            new_ctx[self.n_var.name] = binding
            result = self.n_expr.evaluate(mh, new_ctx)
            assert isinstance(result.value, bool)
            if not result.value:
                ok  = False
                loc = binding.location
                break

        return Value(location = loc,
                     value    = ok,
                     typ      = self.typ)


##############################################################################
# AST Nodes (Entities)
##############################################################################

class Entity(Node):
    """Base class for all entities.

    An entity is a concrete object (with a name) for which we need to
    allocate memory. Examples of entities are types and record
    objects.

    :attribute name: unqualified name of the entity
    :type: str

    """
    def __init__(self, name, location):
        super().__init__(location)
        assert isinstance(name, str)
        self.name = name


class Quantified_Variable(Entity):
    """Variable used in quantified expression.

    A quantified expression declares and binds a variable, for which
    we need a named entity. For example in::

      (forall x in array => x > 1)
              ^

    We represent this first x as a :class:`Quantified_Variable`, the
    second x will be an ordinary :class:`Name_Reference`.

    :attribute typ: type of the variable (i.e. element type of the array)
    :type: Type

    """
    def __init__(self, name, location, typ):
        super().__init__(name, location)
        assert isinstance(typ, Type)
        self.typ = typ

    def dump(self, indent=0):  # pragma: no cover
        self.write_indent(indent, "Quantified Variable %s" % self.name)
        self.typ.dump(indent + 1)


class Type(Entity):
    """Abstract base class for all types.

    """


class Builtin_Type(Type):
    """Abstract base class for all builtin types.

    """
    LOCATION = Location(file_name = "<builtin>")

    def __init__(self, name):
        super().__init__(name, Builtin_Type.LOCATION)

    def dump(self, indent=0):  # pragma: no cover
        self.write_indent(indent, self.__class__.__name__)


class Builtin_Function(Entity):
    """Builtin functions.

    These are auto-generated by the :class:`~trlc.trlc.Source_Manager`.

    :attribute arity: number of parameters
    :type: int
    """
    LOCATION = Location(file_name = "<builtin>")

    def __init__(self, name, arity):
        super().__init__(name, Builtin_Function.LOCATION)
        assert isinstance(arity, int)
        assert arity >= 0
        self.arity = arity

    def dump(self, indent=0):  # pragma: no cover
        self.write_indent(indent, self.__class__.__name__ + " " + self.name)


class Array_Type(Type):
    """Anonymous array type.

    These are declared implicitly for each record component that has
    an array specifier::

      foo Integer [5 .. *]
                  ^

    :attribute lower_bound: minimum number of elements
    :type: int

    :attribute upper_bound: maximum number of elements (or None)
    :type: int

    :attribute element_type: type of the array elements
    :type: Type

    """
    def __init__(self, location, element_type, lower_bound, upper_bound):
        assert isinstance(element_type, Type)
        assert isinstance(lower_bound, int)
        assert lower_bound >= 0
        assert upper_bound is None or isinstance(upper_bound, int)
        assert upper_bound is None or upper_bound >= 0

        if upper_bound is None:
            if lower_bound == 0:
                name = "array of %s" % element_type.name
            else:
                name = "array of at least %u %s" % (lower_bound,
                                                    element_type.name)
        elif lower_bound == upper_bound:
            name = "array of %u %s" % (lower_bound,
                                       element_type.name)
        else:
            name = "array of %u to %u %s" % (lower_bound,
                                             upper_bound,
                                             element_type.name)
        super().__init__(name, location)
        self.lower_bound  = lower_bound
        self.upper_bound  = upper_bound
        self.element_type = element_type


class Builtin_Integer(Builtin_Type):
    """Builtin integer type."""
    def __init__(self):
        super().__init__("Integer")


class Builtin_Boolean(Builtin_Type):
    """Builtin boolean type."""
    def __init__(self):
        super().__init__("Boolean")


class Builtin_String(Builtin_Type):
    """Builtin string type."""
    def __init__(self):
        super().__init__("String")


class Package(Entity):
    """Packages.

    A package is declared when it is first encountered (in either a
    rsl or trlc file). A package contains all symbols declared in it,
    both types and record objects. A package is not associated with
    just a single file, it can be spread over multiple files.

    :attribute symbols: symbol table of the package
    :type: Symbol_Table


    """
    def __init__(self, name, location, builtin_stab):
        super().__init__(name, location)
        assert isinstance(builtin_stab, Symbol_Table)
        self.symbols = Symbol_Table()
        self.symbols.make_visible(builtin_stab)
        self.declared_late = False

    def dump(self, indent=0):  # pragma: no cover
        self.write_indent(indent, "Package %s" % self.name)
        self.symbols.dump(indent + 1)


class Record_Type(Type):
    """A user-defined record type.

    In this example::

      type T  "optional description of T" extends Root_T {
           ^1 ^2                         ^3

    :attribute parent: root type or None, indicated by (3) above
    :type: Record_Type

    :attribute components: record components (including inherited)
    :type: Symbol_Table[Record_Component]

    :attribute description: user-supplied description of the record or \
    None, see (2)
    :type: str

    :attribute checks: used-defined checks for this record (excluding \
    inherited checks)
    :type: list[Check]

    """
    def __init__(self, name, description, location, parent=None):
        super().__init__(name, location)
        assert isinstance(description, str) or description is None
        assert isinstance(parent, Record_Type) or parent is None
        self.parent      = parent
        self.components  = Symbol_Table(self.parent.components
                                        if self.parent
                                        else None)
        self.description = description
        self.checks      = []

    def dump(self, indent=0):  # pragma: no cover
        self.write_indent(indent, "Record_Type %s" % self.name)
        if self.description:
            self.write_indent(indent + 1, "Description: %s" % self.description)
        if self.parent:
            self.write_indent(indent + 1, "Parent: %s" % self.parent.name)
        self.components.dump(indent + 1, omit_heading=True)
        if self.checks:
            self.write_indent(indent + 1, "Checks")
            for n_check in self.checks:
                n_check.dump(indent + 2)
        else:
            self.write_indent(indent + 1, "Checks: None")

    def all_components(self):
        """Convenience function to get a list of all components.

        :rtype: list[Record_Component]
        """
        if self.parent:
            return self.parent.all_components() + \
                list(self.components.table.values())
        else:
            return list(self.components.table.values())

    def add_check(self, n_check):
        assert isinstance(n_check, Check)
        self.checks.append(n_check)

    def iter_checks(self):
        if self.parent:
            yield from self.parent.iter_checks()
        yield from self.checks


class Record_Component(Entity):
    """Component in a record.

    When declaring a record type, for each component an entity is
    declared::

      type T {
         foo "blah" optional Boolean
         ^1  ^2     ^3       ^4

    :attribute description: optional text (see 2) for this field, or None
    :type: str

    :attribute record: a link back to the containing record; for inherited \
    fields this refers back to the original base record type
    :type: Record_Type

    :attribute typ: type of this component (see 4), but note it could also \
    be an anonymous array type
    :type: Type

    :attribute optional: indicates if the component can be null or not (see 3)
    :type: bool

    """

    def __init__(self, name, description, location, record, typ, optional):
        super().__init__(name, location)
        assert isinstance(description, str) or description is None
        assert isinstance(record, Record_Type)
        assert isinstance(typ, Type)
        assert isinstance(optional, bool)
        self.description = description
        self.record      = record
        self.typ         = typ
        self.optional    = optional

    def dump(self, indent=0):  # pragma: no cover
        self.write_indent(indent, "Record_Component %s" % self.name)
        if self.description:
            self.write_indent(indent + 1, "Description: %s" % self.description)
        self.write_indent(indent + 1, "Optional: %s" % self.optional)
        self.write_indent(indent + 1, "Type: %s" % self.typ.name)


class Enumeration_Type(Type):
    """User-defined enumeration types.

    For example::

      enum T  "potato" {
           ^1 ^2

    :attribute description: user supplied optional description, or None
    :type: str

    :attribute literals: the literals in this enumeration
    :type: Symbol_Table[Enumeration_Literal_Spec]

    """
    def __init__(self, name, description, location):
        super().__init__(name, location)
        assert isinstance(description, str) or description is None
        self.literals    = Symbol_Table()
        self.description = description

    def dump(self, indent=0):  # pragma: no cover
        self.write_indent(indent, "Enumeration_Type %s" % self.name)
        if self.description:
            self.write_indent(indent + 1, "Description: %s" % self.description)
        self.literals.dump(indent + 1, omit_heading=True)


class Enumeration_Literal_Spec(Entity):
    """Declared literal in an enumeration declaration.

    Note that for literals mentioned later in record object
    declarations, we use :class:`Enumeration_Literal`. Literal specs
    are used here::

      enum ASIL {
         QM "not safety related"
         ^1 ^2

    :attribute description: the optional user-supplied description, or None
    :type: str

    :attribute enum: a link back to the declaring enumeration
    :type: Enumeration_Type

    """
    def __init__(self, name, description, location, enum):
        super().__init__(name, location)
        assert isinstance(description, str) or description is None
        assert isinstance(enum, Enumeration_Type)
        self.description = description
        self.enum        = enum

    def dump(self, indent=0):  # pragma: no cover
        self.write_indent(indent, "Enumeration_Literal_Spec %s" % self.name)
        if self.description:
            self.write_indent(indent + 1, "Description: %s" % self.description)


class Record_Object(Entity):
    """A declared instance of a record type.

    This is going to be the bulk of all entities created by TRLC::

       section "Potato" {
               ^5
         Requirement PotatoReq {
         ^1          ^2
             component1 = 42
             ^3           ^4

    Note that the name of the object is provided by the name attribute
    of the :class:`Entity` base class.

    :attribute e_typ: the type of the object (see 1)
    :type: Record_Type

    :attribute field: the specific values for all components (see 3 and 4)
    :type: dict[str] Expression

    :attribute section: None or the section this record is contained in (see 5)
    :type: Section

    The actual type of expressions in the field attribute are limited
    to:

    * :class:`Literal`
    * :class:`Array_Aggregate`
    * :class:`Record_Reference`
    * :class:`Implicit_Null`

    """
    def __init__(self, name, location, e_typ, section):
        super().__init__(name, location)
        assert isinstance(e_typ, Record_Type)
        assert isinstance(section, Section) or section is None
        self.e_typ   = e_typ
        self.field   = {name: None
                        for name in self.e_typ.components.all_names()}
        self.section = section

    def to_python_dict(self):
        """Return an evaluated and simplified object for Python.

        For example it might provide::

          {"foo" : [1, 2, 3],
           "bar" : None,
           "baz" : "value"}

        This is a function especially designed for the Python API. The
        name of the object itself is not in this returned dictionary.

        """
        return {name: value.to_python_object()
                for name, value in self.field.items()}

    def assign(self, component, value):
        assert isinstance(component, Record_Component)
        assert isinstance(value, (Literal,
                                  Array_Aggregate,
                                  Record_Reference,
                                  Implicit_Null))
        self.field[component.name] = value

    def dump(self, indent=0):  # pragma: no cover
        self.write_indent(indent, "Record_Object %s" % self.name)
        self.write_indent(indent + 1, "Type: %s" % self.e_typ.name)
        for key, value in self.field.items():
            self.write_indent(indent + 1,
                              "Field %s: %s" % (key,
                                                repr(value.to_string())))
        if self.section:
            self.section.dump(indent + 1)

    def resolve_references(self, mh):
        assert isinstance(mh, Message_Handler)
        for val in self.field.values():
            val.resolve_references(mh)

    def perform_checks(self, mh):
        assert isinstance(mh, Message_Handler)

        ok = True
        for check in self.e_typ.iter_checks():
            # Prints messages, if applicable. Raises exception on
            # fatal checks, which causes this to abort.
            if not check.perform(mh, self):
                ok = False

        return ok


class Section(Entity):
    """A section for readability

    This represents a section construct in TRLC files to group record
    objects together::

      section "Foo" {
              ^^^^^ parent section
         section "Bar" {
                 ^^^^^ section

    :attribute parent: the parent section or None
    :type: Section

    """
    def __init__(self, name, location, parent):
        super().__init__(name, location)
        assert isinstance(parent, Section) or parent is None
        self.parent  = parent

    def dump(self, indent=0):  # pragma: no cover
        self.write_indent(indent, "Section %s" % self.name)
        if self.parent is None:
            self.write_indent(indent + 1, "Parent: None")
        else:
            self.write_indent(indent + 1, "Parent: %s" % self.parent.name)


##############################################################################
# Symbol Table & Scopes
##############################################################################

class Symbol_Table:
    """ Symbol table mapping names to entities
    """
    def __init__(self, parent=None):
        assert isinstance(parent, Symbol_Table) or parent is None
        self.parent   = parent
        self.imported = []
        self.table    = {}

    def all_names(self):
        """ All names in the symbol table

        :rtype: set[str]
        """
        rv = set(self.table)
        if self.parent:
            rv |= self.parent.all_names()
        return rv

    def iter_record_objects(self):
        """ Iterate over all record objects

        :rtype: iterable[Record_Object]
        """
        for item in self.table.values():
            if isinstance(item, Package):
                yield from item.symbols.iter_record_objects()

            elif isinstance(item, Record_Object):
                yield item

    def values(self, subtype=None):
        assert subtype is None or isinstance(subtype, type)
        if self.parent:
            yield from self.parent.values(subtype)
        for name in sorted(self.table):
            if subtype is None or isinstance(self.table[name], subtype):
                yield self.table[name]

    def make_visible(self, stab):
        assert isinstance(stab, Symbol_Table)
        self.imported.append(stab)

    def register(self, mh, entity):
        assert isinstance(mh, Message_Handler)
        assert isinstance(entity, Entity)

        if self.contains(entity.name):
            pdef = self.lookup_direct(mh, entity.name, entity.location)
            mh.error(entity.location,
                     "duplicate definition, previous definition at %s" %
                     pdef.location.to_string(include_column=False))

        else:
            self.table[entity.name] = entity

    def contains(self, name):
        """ Tests if the given name is in the table

        :param name: the name to test
        :type name: str
        :rtype: bool
        """
        assert isinstance(name, str)
        if name in self.table:
            return True
        elif self.parent:
            return self.parent.contains(name)

        for stab in self.imported:
            if stab.contains(name):
                return True

        return False

    def lookup_direct(self, mh, name, error_location, required_subclass=None):
        """Retrieve an object from the table

        For example::

          pkg = stab.lookup_direct(mh,
                                   "potato",
                                   Location("foobar.txt", 42),
                                   Package)

        This would search for an object named ``potato``. If it is
        found, and it is a package, it is returned. If it is not a
        Package, then the following error is issued::

          foobar.txt:42: error: Enumeration_Type potato is not a Package

        If it is not found at all, then the following error is issued::

          foobar.txt:42: error: unknown symbol potato

        :param mh: The message handler to use
        :type mh: Message_Handler

        :param name: The name to search for
        :type name: str

        :param error_location: Where to create the error if the name is \
        not found
        :type error_location: Location

        :param required_subclass: If set, creates an error if the object \
        is not an instance of the given class
        :type required_subclass: type

        :raise TRLC_Error: if the name is not in the table
        :raise TRLC_Error: if the object is not of the required subclass
        :returns: the specified entity
        :rtype: Entity

        """
        assert isinstance(mh, Message_Handler)
        assert isinstance(name, str)
        assert isinstance(error_location, Location)
        assert isinstance(required_subclass, type) or required_subclass is None

        ptr     = self
        options = []
        for ptr in [self] + self.imported:
            while ptr:
                if name in ptr.table:
                    rv = ptr.table[name]
                    if required_subclass is not None and \
                       not isinstance(rv, required_subclass):
                        mh.error(error_location,
                                 "%s %s is not a %s" %
                                 (rv.__class__.__name__,
                                  name,
                                  required_subclass.__name__))
                    return rv
                else:
                    options += list(ptr.table)
                    ptr      = ptr.parent

        matches = get_close_matches(
            word          = name,
            possibilities = options,
            n             = 1)
        if matches:
            mh.error(error_location,
                     "unknown symbol %s, did you mean %s?" %
                     (name,
                      matches[0]))
        else:
            mh.error(error_location,
                     "unknown symbol %s" % name)

    def lookup(self, mh, referencing_token, required_subclass=None):
        assert isinstance(mh, Message_Handler)
        assert isinstance(referencing_token, Token)
        assert referencing_token.kind in ("IDENTIFIER", "BUILTIN")
        assert isinstance(required_subclass, type) or required_subclass is None

        return self.lookup_direct(
            mh                = mh,
            name              = referencing_token.value,
            error_location    = referencing_token.location,
            required_subclass = required_subclass)

    def write_indent(self, indent, message):
        assert isinstance(indent, int)
        assert indent >= 0
        assert isinstance(message, str)
        print(" " * (3 * indent) + message)

    def dump(self, indent=0, omit_heading=False):  # pragma: no cover
        if omit_heading:
            new_indent = indent
        else:
            self.write_indent(indent, "Symbol_Table")
            new_indent = indent + 1
        ptr = self
        while ptr:
            for name in sorted(ptr.table):
                ptr.table[name].dump(new_indent)
            ptr = ptr.parent

    @classmethod
    def create_global_table(cls, mh):
        stab = Symbol_Table()
        stab.register(mh, Builtin_Integer())
        stab.register(mh, Builtin_Boolean())
        stab.register(mh, Builtin_String())
        stab.register(mh,
                      Builtin_Function("trlc:len", 1))
        stab.register(mh,
                      Builtin_Function("trlc:startswith", 2))
        stab.register(mh,
                      Builtin_Function("trlc:endswith", 2))
        stab.register(mh,
                      Builtin_Function("trlc:matches", 2))

        return stab


class Scope:
    def __init__(self):
        self.scope = []

    def push(self, stab):
        assert isinstance(stab, Symbol_Table)
        self.scope.append(stab)

    def pop(self):
        self.scope.pop()

    def contains(self, name):
        assert isinstance(name, str)

        for stab in reversed(self.scope):
            if stab.contains(name):
                return True
        return False

    def lookup(self, mh, referencing_token, required_subclass=None):
        assert len(self.scope) >= 1
        assert isinstance(mh, Message_Handler)
        assert isinstance(referencing_token, Token)
        assert referencing_token.kind in ("IDENTIFIER", "BUILTIN")
        assert isinstance(required_subclass, type) or required_subclass is None

        for stab in reversed(self.scope[1:]):
            if stab.contains(referencing_token.value):
                return stab.lookup(mh, referencing_token, required_subclass)
        return self.scope[0].lookup(mh, referencing_token, required_subclass)

    def size(self):
        return len(self.scope)
